#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import asyncio
import importlib.util
import importlib.metadata
import logging
import os
import shtab
import subprocess
import sys
import threading
import time
from typing import Optional

# Global Constants and Logging
HOME_DIR = os.path.expanduser("~")
CONFIG_DIR = ".agefreighter"
COMPLETION_FILE = "_agefreighter.completion"

logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def ensure_dir(path: str) -> None:
    """Ensure that a directory exists."""
    if not os.path.exists(path):
        os.makedirs(path)


def get_current_shell():
    ppid = os.getppid()  # Get parent process ID
    try:
        # Use ps to get the command name of the parent process.
        proc = subprocess.run(
            ["ps", "-p", str(ppid), "-o", "comm="],
            capture_output=True,
            text=True,
            check=True,
        )
        shell_path = proc.stdout.strip()
        if "bash" in shell_path:
            return "bash"
        elif "zsh" in shell_path:
            return "zsh"
        elif "tcsh" in shell_path:
            return "tcsh"
        else:
            return os.path.basename(shell_path)
    except subprocess.CalledProcessError:
        return None


def generate_completion(parser: argparse.ArgumentParser) -> None:
    """Generate the shell completion script and write it to a file."""
    shell = get_current_shell()
    if not shell:
        log.error("Failed to detect shell. Please set the SHELL environment variable.")
        sys.exit(1)
    completion_script = shtab.complete(parser, shell=shell)

    ensure_dir(os.path.join(HOME_DIR, CONFIG_DIR))
    completion_path = os.path.join(HOME_DIR, CONFIG_DIR, COMPLETION_FILE)

    print(f"Writing completion script to {completion_path}")
    with open(completion_path, "w") as f:
        f.write(completion_script)

    if shell == "bash":
        lines_to_add = f'source "$HOME/{CONFIG_DIR}/{COMPLETION_FILE}"'
        rc_file = os.path.join(HOME_DIR, ".bashrc")
    elif shell == "zsh":
        lines_to_add = f"""if type agefreighter&>/dev/null; then
    FPATH=~/{CONFIG_DIR}:$FPATH
    autoload -Uz compinit
    compinit
fi"""
        rc_file = os.path.join(HOME_DIR, ".zprofile")
        if not os.path.exists(rc_file):
            rc_file = os.path.join(HOME_DIR, ".zshrc")
    else:
        rc_file = None

    if rc_file and os.path.exists(rc_file):
        with open(rc_file, "a+") as f:
            f.seek(0)
            contents = f.read()
            if lines_to_add not in contents:
                f.write("\n" + lines_to_add + "\n")
                print(f"Added line to {rc_file}")
    print(
        f"Completion script generated successfully.\nPlease execute `source {rc_file}` or restart your shell to enable completion."
    )
    sys.exit(0)


def check_first_run() -> None:
    """Create a marker file on the first run and print a completion hint."""
    clean_old_files()
    ensure_dir(os.path.join(HOME_DIR, CONFIG_DIR))
    marker_file = os.path.join(HOME_DIR, CONFIG_DIR, ".agefreighter_first_run")
    if not os.path.exists(marker_file):
        try:
            with open(marker_file, "w") as f:
                f.write("agefreighter has been executed.")
        except Exception as e:
            log.error(f"Failed to create first run marker file: {e}")
        print(
            "If you'd like to enable completion for AGEFreighter, please execute `agefreighter --generate-completion`."
        )


# since 1.0.0a10
def clean_old_files() -> None:
    for file in [".g2c_cache", ".agefreighter_first_run"]:
        if os.path.exists(os.path.join(HOME_DIR, file)):
            os.remove(os.path.join(HOME_DIR, file))


def check_and_install(package_name: str, module_name: Optional[str] = None) -> None:
    """
    Ensure a module is available. If not, prompt to install it via pip or uv.

    :param package_name: The package name to install.
    :param module_name: The module name to check (defaults to package_name).
    """
    module_name = module_name or package_name
    try:
        spec = importlib.util.find_spec(module_name)
    except ModuleNotFoundError:
        spec = None

    if spec is not None:
        log.debug(f"Module '{module_name}' is already available.")
        return

    answer = (
        input(f"Required module '{module_name}' is not installed. Install it? [Y/n]: ")
        .strip()
        .lower()
    )
    if answer not in ("", "y", "yes"):
        print(f"'{package_name}' installation was denied. Exiting subcommand.")
        sys.exit(1)

    install_cmd = [sys.executable, "-m", "pip", "install", package_name]
    try:
        subprocess.run(install_cmd, capture_output=True, text=True, check=True)
    except subprocess.CalledProcessError as e:
        if "No module named pip" in (e.stderr or ""):
            print("pip module is not available. Trying with uv...")
            alt_install_cmd = ["uv", "add", "--dev", package_name]
            try:
                subprocess.run(alt_install_cmd, check=True)
            except subprocess.CalledProcessError:
                print(f"Failed to install '{package_name}' using uv.")
                sys.exit(1)
        else:
            print(f"Failed to install '{package_name}'.")
            sys.exit(1)


def run_flask(flask_port: int, log_level: int = logging.INFO) -> None:
    """Run the Flask app on the specified port."""
    from agefreighter.view import app

    app.logger.setLevel(log_level)
    app.run(port=flask_port)


def create_parser() -> argparse.ArgumentParser:
    """Create and return the argument parser with all options and subcommands."""
    parser = argparse.ArgumentParser(
        description="AGEFreighter, a tool to export data from various sources and load it into Apache AGE."
    )
    # Global arguments
    parser.add_argument(
        "--graphname",
        type=str,
        default="FROM_AGEFREIGHTER",
        help="Name of a new graph on Apache AGE",
    )
    parser.add_argument(
        "--pg-con-str",
        type=str,
        default=os.environ.get("PG_CONNECTION_STRING", ""),
        help="Connection string of the Azure Database for PostgreSQL",
    )
    parser.add_argument(
        "--pg-min-connections",
        type=int,
        default=4,
        help="Minimum number of connections to PostgreSQL",
    )
    parser.add_argument(
        "--pg-max-connections",
        type=int,
        default=64,
        help="Maximum number of connections to PostgreSQL",
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="Enable debug logging"
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"agefreighter {importlib.metadata.version('agefreighter')}",
        help="Show version information",
    )
    parser.add_argument(
        "--generate-completion",
        action="store_true",
        help="Generate the completion script and exit",
    )

    # Subparsers
    subparsers = parser.add_subparsers(dest="subparser", required=True)

    # load subcommand
    parser_load = subparsers.add_parser("load", help="Load data into Apache AGE")
    parser_load.add_argument(
        "--source-type",
        type=str,
        choices=["neo4j", "cosmosdb", "pgsql", "csv"],
        default="neo4j",
        help="Source type of the graph data",
    )
    parser_load.add_argument(
        "--trial",
        action="store_true",
        default=False,
        help="Extract only 100 edges per relationship type",
    )
    parser_load.add_argument(
        "--save-temps",
        action="store_true",
        default=False,
        help="Save data from source as CSV files into a directory",
    )
    parser_load.add_argument(
        "--chunk-size", type=int, default=10000, help="Chunk size for exporting data"
    )
    parser_load.add_argument(
        "--progress", action="store_true", default=True, help="Show progress"
    )
    parser_load.add_argument(
        "--config", type=str, default="", help="Path to the configuration file"
    )
    parser_load.add_argument(
        "--neo4j-uri",
        type=str,
        default=os.environ.get("NEO4J_URI", ""),
        help="Neo4j URI",
    )
    parser_load.add_argument(
        "--neo4j-user",
        type=str,
        default=os.environ.get("NEO4J_USER", ""),
        help="Neo4j username",
    )
    parser_load.add_argument(
        "--neo4j-password",
        type=str,
        default=os.environ.get("NEO4J_PASSWORD", ""),
        help="Neo4j password",
    )
    parser_load.add_argument(
        "--neo4j-database", type=str, default="", help="Neo4j database"
    )
    parser_load.add_argument(
        "--cosmos-endpoint",
        type=str,
        default=os.environ.get("COSMOS_ENDPOINT", ""),
        help="Cosmos endpoint",
    )
    parser_load.add_argument(
        "--cosmos-key",
        type=str,
        default=os.environ.get("COSMOS_KEY", ""),
        help="Cosmos key",
    )
    parser_load.add_argument(
        "--cosmos-database",
        type=str,
        default=os.environ.get("COSMOS_DATABASE", ""),
        help="Cosmos database",
    )
    parser_load.add_argument(
        "--cosmos-container",
        type=str,
        default=os.environ.get("COSMOS_CONTAINER", ""),
        help="Cosmos container",
    )
    parser_load.add_argument(
        "--src-pg-con-str",
        type=str,
        default=os.environ.get("SRC_PG_CONNECTION_STRING", ""),
        help="Source PostgreSQL connection string",
    )

    # view subcommand
    parser_view = subparsers.add_parser("view", help="View data in Apache AGE")
    parser_view.add_argument(
        "--flask-port", type=int, default=5000, help="Port to run the server on"
    )

    # parse subcommand
    parser_parse = subparsers.add_parser("parse", help="Parse a cypher query")
    parser_parse.add_argument(
        "cypher_query",
        type=str,
        default="MATCH (n) RETURN n",
        help="Cypher query to be parsed",
    )

    # generate subcommand
    parser_generate = subparsers.add_parser("generate", help="Generate dummy data")
    parser_generate.add_argument(
        "--pattern-no", type=int, default=1, help="Pattern number to generate"
    )
    parser_generate.add_argument(
        "--multiplier",
        type=int,
        default=1,
        help="Multiplier for the number of nodes and edges",
    )

    # convert subcommand
    parser_convert = subparsers.add_parser(
        "convert",
        help="Convert Gremlin / Cypher queries to Cypher queries for Apache AGE.",
    )
    parser_convert.add_argument(
        "-l",
        "--query-language",
        choices=["gremlin", "cypher"],
        default="gremlin",
        help="Source language of the query",
    )
    parser_convert.add_argument(
        "-k",
        "--openai-api-key",
        type=str,
        default=os.environ.get("OPENAI_API_KEY", ""),
        help="OpenAI API key to use.",
    )
    parser_convert.add_argument(
        "-m", "--model", type=str, default="gpt-4o-mini", help="OpenAI model to use."
    )
    parser_convert.add_argument(
        "-d",
        "--dryrun",
        action="store_true",
        default=False,
        help="Dry run with PostgreSQL.",
    )
    parser_convert.add_argument(
        "--pg-con-str-for-dryrun",
        type=str,
        default=os.environ.get("PG_CONNECTION_STRING", ""),
        help="Connection string of the Azure Database for PostgreSQL",
    )
    parser_convert.add_argument(
        "--graph-for-dryrun",
        type=str,
        default="GRAPH_FOR_DRYRUN",
        help="Graph name for dry run with PostgreSQL.",
    )
    group_convert = parser_convert.add_mutually_exclusive_group(required=True)
    group_convert.add_argument(
        "-q",
        "--query",
        type=str,
        default="GRAPH_FOR_DRYRUN",
        help="Graph name for dry run with PostgreSQL.",
    )
    group_convert.add_argument(
        "-f",
        "--filepath",
        type=str,
        help="Path to the source code file (.py, .java, .cs, .txt)",
    )
    group_convert.add_argument(
        "-u",
        "--url",
        type=str,
        help="URL to the source code file (.py, .java, .cs, .txt)",
    )

    # prepare subcommand
    parser_prepare = subparsers.add_parser(
        "prepare", help="Prepare data for testing AGEFreighter."
    )
    parser_prepare.add_argument(
        "--target-type",
        type=str,
        choices=["neo4j", "cosmosdb", "pgsql"],
        default="neo4j",
        help="Targeted type of the source of graph data.",
    )
    parser_prepare.add_argument(
        "--data-dir",
        type=str,
        default="docs",
        help="Directory containing the data files",
    )
    parser_prepare.add_argument(
        "--base-file",
        type=str,
        default="customer_product_bought.csv",
        help="Base file name for the data files",
    )
    parser_prepare.add_argument(
        "--neo4j-uri",
        type=str,
        default=os.environ.get("NEO4J_URI", ""),
        help="Neo4j URI",
    )
    parser_prepare.add_argument(
        "--neo4j-user",
        type=str,
        default=os.environ.get("NEO4J_USER", ""),
        help="Neo4j username",
    )
    parser_prepare.add_argument(
        "--neo4j-password",
        type=str,
        default=os.environ.get("NEO4J_PASSWORD", ""),
        help="Neo4j password",
    )
    parser_prepare.add_argument(
        "--neo4j-database", type=str, default="", help="Neo4j database"
    )
    parser_prepare.add_argument(
        "--cosmos-gremlin-endpoint",
        type=str,
        default=os.environ.get("COSMOS_GREMLIN_ENDPOINT", ""),
        help="Cosmos Gremlin endpoint",
    )
    parser_prepare.add_argument(
        "--cosmos-key",
        type=str,
        default=os.environ.get("COSMOS_KEY", ""),
        help="Cosmos key",
    )
    parser_prepare.add_argument(
        "--cosmos-database",
        type=str,
        default=os.environ.get("COSMOS_DATABASE", ""),
        help="Cosmos database",
    )
    parser_prepare.add_argument(
        "--cosmos-container",
        type=str,
        default=os.environ.get("COSMOS_CONTAINER", ""),
        help="Cosmos container",
    )
    parser_prepare.add_argument(
        "--src-pg-con-str",
        type=str,
        default=os.environ.get("SRC_PG_CONNECTION_STRING", ""),
        help="Source PostgreSQL connection string",
    )

    return parser


def parse_arguments() -> argparse.Namespace:
    """Parse command line arguments and trigger completion generation if needed."""
    parser = create_parser()
    if "--generate-completion" in sys.argv:
        generate_completion(parser)
    return parser.parse_args()


# Subcommand Handlers
async def handle_load(args) -> None:
    match args.source_type:
        case "neo4j":
            if not (args.neo4j_uri and args.neo4j_user and args.neo4j_password):
                log.error(
                    "NEO4J_URI, NEO4J_USER, and NEO4J_PASSWORD not set. Set via environment variable or argument."
                )
                sys.exit(1)
            check_and_install("neo4j")
            from agefreighter.neo4jexporter import Neo4jExporter

            try:
                async with Neo4jExporter(
                    dsn=args.pg_con_str,
                    min_connections=args.pg_min_connections,
                    max_connections=args.pg_max_connections,
                    uri=args.neo4j_uri,
                    user=args.neo4j_user,
                    password=args.neo4j_password,
                    database=args.neo4j_database,
                    trial=args.trial,
                    save_temps=args.save_temps,
                    progress=args.progress,
                    graph_name=args.graphname,
                    chunk_size=args.chunk_size,
                    log_level=logging.DEBUG if args.debug else logging.INFO,
                ) as exporter:
                    await exporter.export()
                    await exporter.copy()
            except Exception as e:
                log.error("An error occurred during export: %s", e)
                sys.exit(1)
        case "csv":
            if not args.config:
                log.error(
                    "Config file is required for CSV export. See samples in configs directory."
                )
                sys.exit(1)
            if not os.path.exists(os.path.abspath(args.config)):
                log.error("Config file '%s' does not exist.", args.config)
                sys.exit(1)
            from agefreighter.csvexporter import CSVExporter

            try:
                async with CSVExporter(
                    dsn=args.pg_con_str,
                    min_connections=args.pg_min_connections,
                    max_connections=args.pg_max_connections,
                    config=os.path.abspath(args.config),
                    trial=args.trial,
                    save_temps=args.save_temps,
                    progress=args.progress,
                    graph_name=args.graphname,
                    chunk_size=args.chunk_size,
                    log_level=logging.DEBUG if args.debug else logging.INFO,
                ) as exporter:
                    await exporter.export()
                    await exporter.copy()
            except Exception as e:
                log.error("An error occurred during export: %s", e)
                sys.exit(1)
        case "cosmosdb":
            if not (
                args.cosmos_endpoint
                and args.cosmos_key
                and args.cosmos_database
                and args.cosmos_container
            ):
                log.error(
                    "COSMOS_ENDPOINT, COSMOS_KEY, COSMOS_DATABASE, and COSMOS_CONTAINER not set. Set via environment variable or argument."
                )
                sys.exit(1)
            check_and_install("azure-cosmos", "azure.cosmos")
            from agefreighter.cosmosnosqlexporter import CosmosNoSQLExporter

            try:
                async with CosmosNoSQLExporter(
                    dsn=args.pg_con_str,
                    min_connections=args.pg_min_connections,
                    max_connections=args.pg_max_connections,
                    cosmos_endpoint=args.cosmos_endpoint,
                    cosmos_key=args.cosmos_key,
                    cosmos_database=args.cosmos_database,
                    cosmos_container=args.cosmos_container,
                    trial=args.trial,
                    save_temps=args.save_temps,
                    progress=args.progress,
                    graph_name=args.graphname,
                    chunk_size=args.chunk_size,
                    log_level=logging.DEBUG if args.debug else logging.INFO,
                ) as exporter:
                    await exporter.export()
                    await exporter.copy()
            except Exception as e:
                log.error("An error occurred during export: %s", e)
                sys.exit(1)
        case "pgsql":
            if not args.src_pg_con_str:
                log.error(
                    "SRC_PG_CON_STR not set. Set via environment variable or argument."
                )
                sys.exit(1)
            if not args.config:
                log.error(
                    "Config file is required for PostgreSQL export. See samples in configs directory."
                )
                sys.exit(1)
            if not os.path.exists(os.path.abspath(args.config)):
                log.error("Config file '%s' does not exist.", args.config)
                sys.exit(1)
            from agefreighter.pgsqlexporter import PGSQLExporter

            try:
                async with PGSQLExporter(
                    dsn=args.pg_con_str,
                    min_connections=args.pg_min_connections,
                    max_connections=args.pg_max_connections,
                    src_dsn=args.src_pg_con_str,
                    config=os.path.abspath(args.config),
                    trial=args.trial,
                    save_temps=args.save_temps,
                    progress=args.progress,
                    graph_name=args.graphname,
                    chunk_size=args.chunk_size,
                    log_level=logging.DEBUG if args.debug else logging.INFO,
                ) as exporter:
                    await exporter.export()
                    await exporter.copy()
            except Exception as e:
                log.error("An error occurred during export: %s", e)
                sys.exit(1)
        case _:
            log.error("Source type '%s' is not implemented.", args.source_type)
            sys.exit(1)


def handle_view(args) -> None:
    check_and_install("flask")
    check_and_install("ply")
    flask_thread = threading.Thread(
        target=run_flask,
        args=(args.flask_port, logging.DEBUG if args.debug else logging.INFO),
    )
    flask_thread.daemon = True
    flask_thread.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Received interrupt. Exiting...")


def handle_parse(args) -> None:
    if not args.cypher_query:
        log.error("Cypher query not provided.")
        sys.exit(1)
    check_and_install("ply")
    from agefreighter.cypherparser import CypherParser

    parser_obj = CypherParser(log_level=logging.DEBUG if args.debug else logging.INFO)
    print(parser_obj.parse(args.cypher_query))


async def handle_generate(args) -> None:
    check_and_install("faker")
    import agefreighter.generator as generator

    await generator.main(
        pattern_no=args.pattern_no,
        multiplier=args.multiplier,
        log_level=logging.DEBUG if args.debug else logging.INFO,
    )


def handle_convert(args) -> None:
    if args.query_language == "gremlin":
        if not args.openai_api_key:
            log.error(
                "OPENAI_API_KEY not set. Set via environment variable or argument."
            )
            sys.exit(1)
        check_and_install("openai")

    check_and_install("ply")
    from agefreighter.converter import ConverterController

    controller = ConverterController(
        query_language=args.query_language,
        api_key=args.openai_api_key,
        model=args.model,
        dryrun=args.dryrun,
        dsn=args.pg_con_str_for_dryrun,
        graph_name=args.graph_for_dryrun,
        query=args.query,
        filepath=args.filepath,
        url=args.url,
        log_level=logging.DEBUG if args.debug else logging.INFO,
    )
    controller.process()


async def handle_prepare(args) -> None:
    check_and_install("pandas")
    from agefreighter.csvdatamanager import CsvDataManager

    csv_manager = CsvDataManager(
        data_dir=args.data_dir,
        base_file=args.base_file,
        log_level=logging.DEBUG if args.debug else logging.INFO,
    )
    match args.target_type:
        case "neo4j":
            if not (args.neo4j_uri and args.neo4j_user and args.neo4j_password):
                log.error(
                    "NEO4J_URI, NEO4J_USER, and NEO4J_PASSWORD not set. Set via environment variable or argument."
                )
                sys.exit(1)
            check_and_install("neo4j")
            from agefreighter.neo4jloader import Neo4jLoader

            neo4j_loader = Neo4jLoader(
                csv_manager=csv_manager,
                neo4j_uri=args.neo4j_uri,
                neo4j_user=args.neo4j_user,
                neo4j_password=args.neo4j_password,
                log_level=logging.DEBUG if args.debug else logging.INFO,
            )
            await neo4j_loader.load_data()
        case "pgsql":
            from agefreighter.pgsqlloader import PgsqlLoader

            pgsql_loader = PgsqlLoader(
                csv_manager=csv_manager,
                src_dsn=args.src_pg_con_str,
                log_level=logging.DEBUG if args.debug else logging.INFO,
            )
            await pgsql_loader.load_data()
        case "cosmosdb":
            if not (
                args.cosmos_gremlin_endpoint
                and args.cosmos_key
                and args.cosmos_database
                and args.cosmos_container
            ):
                log.error(
                    "COSMOS_GREMLIN_ENDPOINT, COSMOS_KEY, COSMOS_DATABASE, and COSMOS_CONTAINER not set. Set via environment variable or argument."
                )
                sys.exit(1)
            check_and_install("gremlinpython")
            from agefreighter.cosmosgremlinloader import CosmosGremlinLoader

            cosmos_loader = CosmosGremlinLoader(
                csv_manager=csv_manager,
                cosmos_gremlin_endpoint=args.cosmos_gremlin_endpoint,
                cosmos_key=args.cosmos_key,
                cosmos_database=args.cosmos_database,
                cosmos_container=args.cosmos_container,
                log_level=logging.DEBUG if args.debug else logging.INFO,
            )
            await cosmos_loader.load_data()
        case _:
            log.error("No valid target type provided.")
            sys.exit(1)


async def async_main() -> None:
    args = parse_arguments()

    # Set logging level
    if args.debug:
        log.setLevel(logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
        log.debug("Debug logging enabled.")
    else:
        log.setLevel(logging.INFO)
        logging.getLogger().setLevel(logging.INFO)

    if args.pg_min_connections > args.pg_max_connections:
        log.error("min_connections cannot be greater than max_connections")
        sys.exit(1)
    if not args.pg_con_str:
        log.error(
            "PG_CONNECTION_STRING not set. Set via environment variable or argument."
        )
        sys.exit(1)

    # Dispatch subcommand
    match args.subparser:
        case "load":
            await handle_load(args)
        case "view":
            handle_view(args)
        case "parse":
            handle_parse(args)
        case "generate":
            await handle_generate(args)
        case "convert":
            handle_convert(args)
        case "prepare":
            await handle_prepare(args)
        case _:
            log.error("No valid subcommand provided.")
            sys.exit(1)


def main() -> None:
    """Entry point: Check first-run marker, set event loop policy, and run the async main."""
    check_first_run()
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    try:
        asyncio.run(async_main())
    except asyncio.CancelledError:
        pass
    except KeyboardInterrupt:
        log.error("KeyboardInterrupt received. Shutting down.")
        sys.exit(1)
    except Exception as e:
        log.error("An unhandled exception occurred: %s", e)
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import os
import re
import logging

from agefreighter.cypherparser import CypherParser

from flask import Flask, jsonify, request, render_template  # type: ignore
from psycopg_pool import ConnectionPool

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

app = Flask(__name__)


class ConnectionStringParser:
    """Utility class to parse connection strings into a dictionary."""

    @staticmethod
    def parse(conn_str: str) -> dict:
        """Convert a connection string into a dictionary."""
        conn_dict = {}
        parts = conn_str.split()
        for part in parts:
            if "=" in part:
                key, value = part.split("=", 1)
                conn_dict[key] = value
        return conn_dict


class CypherQueryFormatter:
    """Utility class for formatting Cypher queries for Apache AGE."""

    @staticmethod
    def format_query(graph_name: str, cypher_query: str) -> str:
        """
        Format the provided Cypher query for Apache AGE.

        Raises:
            ValueError: If the query is unsafe or incorrectly formatted.
        """
        if not CypherQueryFormatter.is_safe_cypher_query(cypher_query):
            raise ValueError("Unsafe query")

        # Append LIMIT 50 if no limit is specified.
        if "limit" not in cypher_query.lower():
            cypher_query += " LIMIT 50"

        returns = CypherQueryFormatter.get_return_values(cypher_query)

        # Check for parameterized query usage.
        if re.findall(r"\$(\w+)", cypher_query):
            raise ValueError("Parameterized query")

        if returns:
            ag_types = ", ".join([f"{r} agtype" for r in returns])
            return f"SELECT * FROM cypher('{graph_name}', $$ {cypher_query} $$) AS ({ag_types});"
        else:
            raise ValueError("No return values specified")

    @staticmethod
    def is_safe_cypher_query(cypher_query: str) -> bool:
        """
        Ensure the Cypher query does not contain dangerous commands.

        Returns:
            bool: True if safe, False otherwise.
        """
        tokens = cypher_query.split()
        unsafe_keywords = ["create", "delete", "set", "remove", "merge"]
        return all(token.lower() not in unsafe_keywords for token in tokens)

    @staticmethod
    def get_return_values(cypher_query: str) -> list:
        parser = CypherParser()
        try:
            result = parser.parse(cypher_query)
        except Exception as e:
            log.error(f"Failed to parse Cypher query: {e}")
            return []

        for op, opr, *_ in result:
            if op == "RETURN":
                return opr

        return []


class DatabaseManager:
    """Class for managing database connections and queries."""

    def __init__(self, log_level: int = logging.INFO):
        log.setLevel(log_level)
        self.pool: ConnectionPool
        self.connection_info: dict = {}

    def connect(self, connection_info: dict) -> None:
        """
        Connect to the database using provided connection information.

        Raises:
            Exception: If the connection fails.
        """
        host = connection_info.get("host", "localhost")
        port = connection_info.get("port", "5432")
        dbname = connection_info.get("dbname", "")
        user = connection_info.get("user", "")
        password = connection_info.get("password", "")

        # Construct the connection string with search_path options.
        conn_str = (
            f"host={host} port={port} dbname={dbname} user={user} password={password} "
            f"options='-c search_path=ag_catalog,\"$user\",public'"
        )

        # Initialize the connection pool and test the connection.
        self.pool = ConnectionPool(conninfo=conn_str)
        assert self.pool is not None
        with self.pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute("SELECT 1;")
                cur.fetchone()
        self.connection_info = connection_info  # 保存接続情報
        logging.info("Database connection successful.")

    def execute_query(self, graph_name: str, cypher_query: str) -> dict:
        """
        Execute a formatted Cypher query and return nodes and edges.

        Raises:
            ValueError: For unsafe or invalid queries.
            Exception: For other unexpected errors.
        """
        if self.pool is None:
            raise ValueError("No database connection. Please connect first.")

        formatted_query = CypherQueryFormatter.format_query(graph_name, cypher_query)
        logging.info(f"Executing query: {formatted_query}")

        with self.pool.connection() as conn:
            with conn.cursor() as cur:
                cur.execute(formatted_query)
                try:
                    results = cur.fetchall()
                except Exception as fetch_error:
                    logging.error(f"Error fetching results: {fetch_error}")
                    results = []

        nodes = []
        edges = []
        # Process each result item to extract nodes and edges.
        for row in results:
            for item in row:
                if item.endswith("::vertex"):
                    try:
                        jsn = json.loads(item.rstrip("::vertex"))
                    except json.JSONDecodeError as e:
                        logging.error(f"JSON decode error for vertex: {e}")
                        continue
                    node = {
                        "id": jsn.get("id"),
                        "label": jsn.get("label"),
                        "properties": jsn.get("properties"),
                    }
                    nodes.append(node)
                elif item.endswith("::edge"):
                    try:
                        jsn = json.loads(item.rstrip("::edge"))
                    except json.JSONDecodeError as e:
                        logging.error(f"JSON decode error for edge: {e}")
                        continue
                    edge = {
                        "id": jsn.get("id"),
                        "label": jsn.get("label"),
                        "source": jsn.get("start_id"),
                        "target": jsn.get("end_id"),
                        "properties": jsn.get("properties"),
                    }
                    edges.append(edge)

        logging.info(f"Query result - Nodes: {nodes}, Edges: {edges}")
        return {"nodes": nodes, "edges": edges}

    def get_graph_info(self) -> list:
        """
        Retrieve information about graphs and their labels from the database.

        Returns:
            list: A list of graph information dictionaries.

        Raises:
            ValueError: If no database connection is established.
            Exception: For other unexpected errors.
        """
        if self.pool is None:
            raise ValueError("No database connection. Please connect first.")

        result: list = []
        with self.pool.connection() as conn:
            with conn.cursor() as cur:
                # Retrieve the list of graphs.
                cur.execute(
                    "SELECT graphid, name, namespace FROM ag_graph ORDER BY name;"
                )
                graphs = cur.fetchall()
                for graph in graphs:
                    graphid, graph_name, namespace = graph
                    # Retrieve label information for each graph.
                    cur.execute(
                        "SELECT name, kind, relation FROM ag_label WHERE graph = %s;",
                        (graphid,),
                    )
                    labels = cur.fetchall()
                    nodes = []
                    edges = []
                    for label in labels:
                        label_name, kind, relation = label
                        # Skip system labels (starting with '_').
                        if not label_name.startswith("_"):
                            try:
                                count_query = f"SELECT COUNT(*) FROM {relation};"
                                cur.execute(count_query)
                                query_result = cur.fetchone()
                                if query_result:
                                    count = int(query_result[0])
                                else:
                                    count = 0
                            except Exception as count_error:
                                logging.warning(
                                    f"Error counting records for relation {relation}: {count_error}"
                                )
                                count = None

                            if kind == "v":
                                nodes.append(
                                    {
                                        "name": label_name,
                                        "relation": relation,
                                        "count": count,
                                    }
                                )
                            elif kind == "e":
                                edges.append(
                                    {
                                        "name": label_name,
                                        "relation": relation,
                                        "count": count,
                                    }
                                )
                    result.append(
                        {
                            "graphid": graphid,
                            "graph_name": graph_name,
                            "namespace": namespace,
                            "nodes": nodes,
                            "edges": edges,
                        }
                    )
        return result


db_manager = DatabaseManager()


@app.route("/")
def index():
    """Render the main page with default connection info."""
    default_conn_str = os.environ.get("PG_CONNECTION_STRING", "")
    default_conn = (
        ConnectionStringParser.parse(default_conn_str) if default_conn_str else {}
    )
    return render_template("index.html", default_conn=default_conn)


@app.route("/api/connect", methods=["POST"])
def connect_db():
    """API endpoint to connect to the database."""
    data = request.get_json()
    connection_info = data.get("connection", {})
    try:
        db_manager.connect(connection_info)
        return jsonify({"message": "Successfully connected to the database!"})
    except Exception as e:
        logging.error(f"Database connection error: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/api/connection_status", methods=["GET"])
def connection_status():
    """API endpoint to return the connection status."""
    status = "connected" if db_manager.pool is not None else "disconnected"
    return jsonify({"status": status, "connection_info": db_manager.connection_info})


@app.route("/api/execute_query", methods=["POST"])
def execute_query_endpoint():
    """API endpoint to execute a Cypher query."""
    data = request.get_json()
    cypher_query = data.get("cypher_query", "")
    graph_name = data.get("graph_name", "")

    if not cypher_query:
        return jsonify({"error": "No Cypher query specified"}), 400

    try:
        result = db_manager.execute_query(graph_name, cypher_query)
        return jsonify(result)
    except ValueError as ve:
        logging.error(f"Query error: {ve}")
        return jsonify({"error": str(ve)}), 400
    except Exception as e:
        logging.error(f"Unexpected error during query execution: {e}")
        return jsonify({"error": str(e)}), 500


@app.route("/api/graph_info", methods=["GET"])
def graph_info():
    """API endpoint to retrieve graph information."""
    try:
        result = db_manager.get_graph_info()
        return jsonify(result)
    except ValueError as ve:
        logging.error(f"Graph info error: {ve}")
        return jsonify({"error": str(ve)}), 400
    except Exception as e:
        logging.error(f"Unexpected error retrieving graph info: {e}")
        return jsonify({"error": str(e)}), 500



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
from typing import Generator

import pandas as pd

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class CsvDataManager:
    """
    Manage CSV file operations, including loading a DataFrame and chunking.
    """

    def __init__(
        self,
        data_dir: str,
        base_file: str,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        if data_dir:
            self.csv_file = os.path.abspath(os.path.join(data_dir, base_file))
        log.info("CsvDataManager initialized")

    def get_dataframe(self) -> pd.DataFrame:
        """
        Read the CSV file and return a DataFrame.
        """
        return pd.read_csv(self.csv_file)

    @staticmethod
    def get_chunks(
        df: pd.DataFrame, chunk_size: int
    ) -> Generator[pd.DataFrame, None, None]:
        """
        Yield chunks of the DataFrame.

        Args:
            df (pd.DataFrame): DataFrame to be chunked.
            chunk_size (int): Number of rows per chunk.

        Yields:
            Generator[pd.DataFrame, None, None]: Chunks of the DataFrame.
        """
        for i in range(0, len(df), chunk_size):
            yield df.iloc[i : i + chunk_size].copy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for exporting data from a Cosmos NoSQL database to CSV files and
loading them using AgeFreighter.
"""

import asyncio
import logging
import sys
from typing import Any, Dict, List, Set, Tuple

from azure.cosmos import CosmosClient  # type: ignore
from .agefreighter import AgeFreighter

logging.getLogger("azure.core").setLevel(logging.WARNING)

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class CosmosNoSQLExporter(AgeFreighter):
    """
    Exports nodes and edges from a Cosmos NoSQL database into CSV files for AGE COPY import.
    """

    def __init__(
        self,
        dsn: str,
        min_connections: int,
        max_connections: int,
        cosmos_endpoint: str,
        cosmos_key: str,
        cosmos_database: str,
        cosmos_container: str,
        trial: bool,
        save_temps: bool,
        progress: bool,
        graph_name: str,
        chunk_size: int,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        super().__init__(
            dsn=dsn,
            min_connections=min_connections,
            max_connections=max_connections,
            save_temps=save_temps,
            progress=progress,
            chunk_size=chunk_size,
            log_level=log_level,
        )
        self.cosmos_endpoint = cosmos_endpoint
        self.cosmos_key = cosmos_key
        self.cosmos_database = cosmos_database
        self.cosmos_container = cosmos_container
        self.batch_size = 100
        self.trial = trial
        self.no_of_edges_trial = 100
        self.graph_name = graph_name
        self.id_maps: Dict[str, Dict[str, int]] = {}

        try:
            self.client = CosmosClient(self.cosmos_endpoint, credential=self.cosmos_key)
            self.database_client = self.client.get_database_client(self.cosmos_database)
            self.container_client = self.database_client.get_container_client(
                self.cosmos_container
            )
            log.info("Connected to Cosmos DB at %s.", self.cosmos_endpoint)
        except Exception as exc:
            log.error("Error connecting to Cosmos DB: %s", exc)
            sys.exit(1)

    async def __aenter__(self) -> "CosmosNoSQLExporter":
        await super().__aenter__()
        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        await super().__aexit__(exc_type, exc_value, traceback)

    async def _fetch_edges(self) -> List[Dict[str, Any]]:
        """Fetch all edge documents from Cosmos DB."""
        query = (
            f"SELECT TOP {self.no_of_edges_trial} * FROM c WHERE c._isEdge = true"
            if self.trial
            else "SELECT * FROM c WHERE c._isEdge = true"
        )
        try:
            edges = await asyncio.to_thread(
                lambda: list(
                    self.container_client.query_items(
                        query=query, enable_cross_partition_query=True
                    )
                )
            )
        except Exception as exc:
            log.exception("Failed to fetch edge documents: %s", exc)
            raise
        log.info("Fetched %d edge documents.", len(edges))
        return edges

    def _batch_ids(self, ids: Set[str]) -> List[List[str]]:
        """Split a set of IDs into batches."""
        ids_list = list(ids)
        return [
            ids_list[i : i + self.batch_size]
            for i in range(0, len(ids_list), self.batch_size)
        ]

    async def _fetch_nodes_batch(self, id_batch: List[str]) -> List[Dict[str, Any]]:
        """Fetch node documents for a batch of IDs."""
        id_list_str = ", ".join(f"'{node_id}'" for node_id in id_batch)
        query = f"SELECT * FROM c WHERE c.id IN ({id_list_str})"
        try:
            nodes = await asyncio.to_thread(
                lambda: list(
                    self.container_client.query_items(
                        query=query, enable_cross_partition_query=True
                    )
                )
            )
        except Exception as exc:
            log.exception("Failed to fetch node batch [%s...]: %s", id_batch[:3], exc)
            raise
        return nodes

    async def _fetch_all_nodes(self, node_ids: Set[str]) -> Dict[str, Dict[str, Any]]:
        """
        Fetch all node documents concurrently and return a mapping
        of node ID to its document.
        """
        batches = self._batch_ids(node_ids)
        tasks = [self._fetch_nodes_batch(batch) for batch in batches]
        try:
            results = await asyncio.gather(*tasks)
        except Exception as exc:
            log.exception("Failed to fetch all node documents: %s", exc)
            raise
        nodes: Dict[str, Dict[str, Any]] = {}
        for batch_nodes in results:
            for node in batch_nodes:
                node_id = node.get("id")
                if node_id:
                    nodes[node_id] = node
        log.info("Fetched %d node documents.", len(nodes))
        return nodes

    def _filter_properties(
        self, document: Dict[str, Any], exclude_keys: List[str]
    ) -> Dict[str, Any]:
        """
        Extract and return properties from the document,
        excluding specified keys.
        """
        try:
            return {
                k: v[0]["_value"] for k, v in document.items() if k not in exclude_keys
            }
        except Exception as exc:
            log.exception(
                "Failed to filter properties from document %s: %s",
                document.get("id", "unknown"),
                exc,
            )
            raise

    async def fetch_documents(
        self,
    ) -> Tuple[Dict[str, Dict[str, Any]], Dict[str, Dict[str, Any]]]:
        """
        Fetch documents from Cosmos NoSQL database, group them into vertices and edges,
        write CSV files, and return file path arguments.

        Returns:
            A tuple of two dictionaries: vertex_args and edge_args.
        """
        try:
            # Fetch edge documents.
            edges = await self._fetch_edges()
        except Exception as exc:
            log.exception("Error during fetching edge documents: %s", exc)
            raise

        # Group edge documents by label and collect unique node IDs.
        grouped_edges: Dict[str, List[Dict[str, Any]]] = {}
        node_ids: Set[str] = set()
        for edge in edges:
            edge_label = edge.get("label")
            if not edge_label:
                log.error("Edge document missing label: %s", edge)
                raise ValueError("Edge document does not have a label.")
            grouped_edges.setdefault(edge_label, []).append(edge)
            if edge.get("_vertexId"):
                node_ids.add(edge["_vertexId"])
            if edge.get("_sink"):
                node_ids.add(edge["_sink"])

        # Fetch node documents concurrently.
        try:
            nodes_map = await self._fetch_all_nodes(node_ids)
        except Exception as exc:
            log.exception("Error during fetching node documents: %s", exc)
            raise

        # Group node documents by their label.
        grouped_vertices: Dict[str, List[Dict[str, Any]]] = {}
        for node in nodes_map.values():
            vertex_label = node.get("label")
            if not vertex_label:
                log.error("Node document missing label: %s", node)
                raise ValueError("Node document does not have a label.")
            grouped_vertices.setdefault(vertex_label, []).append(node)

        vertex_args: Dict[str, Dict[str, Any]] = {}
        # Process vertices.
        for label, documents in grouped_vertices.items():
            try:
                await self.create_label_type(label_type="vertex", value=label)
                first_id = await self.get_first_id(self.graph_name, label)
            except Exception as exc:
                log.exception("Error during vertex setup for label %s: %s", label, exc)
                raise

            nodes = []
            for document in documents:
                try:
                    node = {
                        "id": document["id"],
                        **self._filter_properties(
                            document,
                            exclude_keys=[
                                "label",
                                "id",
                                "pk",
                                "_rid",
                                "_self",
                                "_etag",
                                "_attachments",
                                "_ts",
                            ],
                        ),
                    }
                    nodes.append(node)
                except Exception as exc:
                    log.exception("Error processing node document: %s", exc)
                    continue

            # Deduplicate nodes by original ID.
            unique_nodes = {node["id"]: node for node in nodes}
            nodes = list(unique_nodes.values())
            all_data = [
                {"id": idx + first_id, "properties": node}
                for idx, node in enumerate(nodes)
            ]
            try:
                new_map: Dict[str, int] = {}
                for item in all_data:
                    properties = item.get("properties")
                    if isinstance(properties, dict) and "id" in properties:
                        id_val = item.get("id")
                        if isinstance(id_val, (int, str)):
                            new_map[properties["id"]] = int(id_val)
                        else:
                            log.error(
                                "Unexpected type for item['id']: %s", type(id_val)
                            )
                            raise ValueError(
                                f"Expected item['id'] to be int or str convertible to int, got {type(id_val)}"
                            )
                self.id_maps[label] = new_map
                csv_path = await self.write_csv(label, "v", all_data)
            except Exception as exc:
                log.exception(
                    "Error writing CSV for vertices with label %s: %s", label, exc
                )
                raise

            vertex_args[label] = {
                "csv_path": csv_path.replace("\\", "\\\\"),
                "original_id": "id",
                "next_val": str(len(all_data)),
            }

        edge_args: Dict[str, Dict[str, Any]] = {}
        # Process edges.
        for edge_label, documents in grouped_edges.items():
            try:
                await self.create_label_type(label_type="edge", value=edge_label)
                first_id = await self.get_first_id(self.graph_name, edge_label)
            except Exception as exc:
                log.exception(
                    "Error during edge setup for label %s: %s", edge_label, exc
                )
                raise

            edges_list = []
            for idx, document in enumerate(documents):
                try:
                    new_start_id = self.id_maps[document["_vertexLabel"]][
                        document["_vertexId"]
                    ]
                    new_end_id = self.id_maps[document["_sinkLabel"]][document["_sink"]]
                except KeyError as key_exc:
                    log.warning(
                        "Missing mapping for document %s: %s",
                        document.get("id", "unknown"),
                        key_exc,
                    )
                    continue
                try:
                    edge_item = {
                        "id": idx + first_id,
                        "start_id": new_start_id,
                        "end_id": new_end_id,
                        **self._filter_properties(
                            document,
                            exclude_keys=[
                                "label",
                                "id",
                                "_sink",
                                "_sinkLabel",
                                "_sinkPartition",
                                "_vertexId",
                                "_vertexLabel",
                                "_isEdge",
                                "pk",
                                "_rid",
                                "_self",
                                "_etag",
                                "_attachments",
                                "_ts",
                            ],
                        ),
                    }
                    edges_list.append(edge_item)
                except Exception as exc:
                    log.exception(
                        "Error processing edge document %s: %s",
                        document.get("id", "unknown"),
                        exc,
                    )
                    continue

            try:
                csv_path = await self.write_csv(edge_label, "e", edges_list)
            except Exception as exc:
                log.exception(
                    "Error writing CSV for edges with label %s: %s", edge_label, exc
                )
                raise
            edge_args[edge_label] = {
                "csv_path": csv_path.replace("\\", "\\\\"),
                "original_id": "id",
                "next_val": str(len(all_data)),
            }

        return vertex_args, edge_args

    async def export(self) -> None:
        """
        Main export function to process both nodes and edges.
        """
        try:
            await self.set_up_graph(graph_name=self.graph_name, create_graph=True)
            vertex_args, edge_args = await self.fetch_documents()
        except Exception as exc:
            log.exception("Error during export process: %s", exc)
            raise

        self.vertices = vertex_args
        self.edges = edge_args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from datetime import datetime
import logging
import os
import random
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from faker import Faker

# Configure logging; default to INFO
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

fake = Faker()

# Dedicated thread pool for CSV writing
csv_executor = ThreadPoolExecutor(max_workers=4)

BULK_SUPPORTED_NODE_TYPES = {"Product"}


def get_timestamp() -> Dict[str, Any]:
    dt = fake.date_time_this_year(before_now=True, after_now=False, tzinfo=None)
    return {"available_since": dt, "inserted_at": dt}


def prefixed_props(prefix: str, props: Optional[Dict[str, Any]]) -> Dict[str, Any]:
    if props is None:
        return {}
    return {f"{prefix}{k}": v for k, v in props.items()}


def generate_base58_dummy_data(length: int) -> str:
    base58_chars = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
    return "".join(random.choice(base58_chars) for _ in range(length))


# ------------------ Lightweight Node Generators ------------------
# Standard generators return a dictionary representing a node.
def generate_airport(id: str) -> dict:
    return {
        "id": id,
        "Name": f"{fake.city()} Airport",
        "City": fake.city(),
        "Country": fake.country(),
        "IATA": fake.unique.bothify(text="???").upper(),
        "ICAO": fake.unique.bothify(text="????").upper(),
        "Latitude": fake.latitude(),
        "Longitude": fake.longitude(),
        "Altitude": fake.random_int(min=0, max=1000),
        "Timezone": fake.timezone(),
        "DST": random.choice(["E", "A", "S", "O", "Z", "N", "U"]),
        "Tz": fake.timezone(),
    }


def generate_customer(id: str) -> dict:
    return {
        "id": id,
        "Name": fake.name(),
        "Address": fake.address().replace("\n", " "),
        "Email": fake.email(),
        "Phone": fake.phone_number(),
    }


# Bulk generator for Product nodes using vectorized numeric operations.
def generate_product_bulk(ids: List[str]) -> List[dict]:
    n = len(ids)
    # Vectorized numeric generation:
    # Generate 5-digit numbers then divide by 100 for price in range ~[100, 1000)
    prices = np.random.randint(10000, 100000, size=n) / 100.0
    weights = np.random.randint(100, 1001, size=n)
    sizes = np.random.choice(["S", "M", "L", "XL"], size=n)
    result = []
    for i, id in enumerate(ids):
        result.append(
            {
                "id": id,
                "Phrase": fake.catch_phrase(),
                "SKU": fake.ean13(),
                "Price": round(prices[i], 2),
                "Color": fake.color_name(),
                "Size": sizes[i],
                "Weight": int(weights[i]),
            }
        )
    return result


def generate_person(id: str) -> dict:
    return {"id": id, "Name": fake.name()}


def generate_manufacturer(id: str) -> dict:
    return {
        "id": id,
        "Name": fake.company(),
        "Address": fake.address(),
        "Phone": fake.phone_number(),
        "Email": fake.email(),
        "Website": fake.url(),
    }


def generate_country(id: str) -> dict:
    return {
        "id": id,
        "Name": fake.country(),
        "Capital": fake.city(),
        "Population": fake.random_int(min=1_000_000, max=1_000_000_000),
        "ISO": fake.unique.bothify(text="??").upper(),
        "TLD": fake.unique.bothify(text="???").lower(),
        "FlagURL": fake.image_url(),
    }


def generate_city(id: str) -> dict:
    return {
        "id": id,
        "Name": fake.city(),
        "Latitude": fake.latitude(),
        "Longitude": fake.longitude(),
    }


def generate_bitcoinaddress(id: str) -> dict:
    base_props = get_timestamp()
    extra = {
        "address": random.choice(["cryptoaddress:", "bitcoinaddress:"])
        + generate_base58_dummy_data(16),
        "schema_version": "1",
    }
    base_props.update(extra)
    base_props["id"] = id
    return base_props


def generate_cookie(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "uaid": fake.uuid4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_ip(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "address": fake.ipv4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_phone(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "address": fake.ipv4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_email(id: str) -> dict:
    base_props = get_timestamp()
    email = fake.email()
    handle, domain = email.split("@")
    base_props.update(
        {
            "email": email,
            "domain": domain,
            "handle": handle,
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_payment(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "payment_id": fake.uuid4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_partnerenduser(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "partner_end_user_id": fake.uuid4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_creditcard(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "expiry_month": fake.month(),
            "expiry_year": fake.year(),
            "masked_number": fake.credit_card_number(card_type=None),
            "creditcard_identifier": fake.uuid4(),
            "schema_version": "1",
        }
    )
    base_props["id"] = id
    return base_props


def generate_cryptoaddress(id: str) -> dict:
    base_props = get_timestamp()
    base_props.update(
        {
            "address": generate_base58_dummy_data(16),
            "currency": random.choice(["BTC", "ETH", "LTC", "XRP"]),
            "full_address": generate_base58_dummy_data(32),
            "schema_version": "1",
            "tag": generate_base58_dummy_data(4),
        }
    )
    base_props["id"] = id
    return base_props


NODE_GENERATORS: Dict[str, Any] = {
    "AirPort": generate_airport,
    "Customer": generate_customer,
    # Use the bulk generator for products.
    "Product": generate_product_bulk,
    "Person": generate_person,
    "Manufacturer": generate_manufacturer,
    "Country": generate_country,
    "City": generate_city,
    "BitcoinAddress": generate_bitcoinaddress,
    "Cookie": generate_cookie,
    "IP": generate_ip,
    "Phone": generate_phone,
    "Email": generate_email,
    "Payment": generate_payment,
    "PartnerEndUser": generate_partnerenduser,
    "CreditCard": generate_creditcard,
    "CryptoAddress": generate_cryptoaddress,
}


# ------------------ Lightweight Edge Generators ------------------
def generate_airroute(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    # This function is kept for single-edge generation.
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **prefixed_props("start_", start_props),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **prefixed_props("end_", end_props),
        "distance": fake.random_int(min=100, max=16000),
    }


# Bulk generator for AirRoute edges using vectorized operations.
def generate_airroute_bulk(
    start_idx: int,
    end_idx: int,
    edge_props: Dict[str, Any],
    start_node_data: List[Dict[str, Any]],
    end_node_data: List[Dict[str, Any]],
) -> List[dict]:
    n = end_idx - start_idx
    # Precompute random indices for start and end nodes.
    start_indices = np.random.randint(0, len(start_node_data), size=n)
    end_indices = np.random.randint(0, len(end_node_data), size=n)
    # Vectorized generation of distances.
    distances = np.random.randint(100, 16001, size=n)
    edges = []
    for i in range(n):
        start = start_node_data[start_indices[i]]
        end = end_node_data[end_indices[i]]
        edge = {
            "id": str(start_idx + i + 1),
            "start_id": start["id"],
            "start_vertex_type": edge_props["start"],
            **prefixed_props("start_", start),
            "end_id": end["id"],
            "end_vertex_type": edge_props["end"],
            **prefixed_props("end_", end),
            "distance": int(distances[i]),
        }
        edges.append(edge)
    return edges


def generate_bought(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **(start_props or {}),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **(end_props or {}),
    }


def generate_has(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **prefixed_props("start_", start_props),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **prefixed_props("end_", end_props),
        "since": fake.date_time(),
    }


def generate_performedby(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        "end_id": end_id,
        "end_vertex_type": end_type,
        **get_timestamp(),
    }


def generate_usedby(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        "end_id": end_id,
        "end_vertex_type": end_type,
        **get_timestamp(),
        "schema_version": "1",
    }


def generate_usedin(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        "end_id": end_id,
        "end_vertex_type": end_type,
        **get_timestamp(),
        "schema_version": "1",
    }


def generate_produce(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **(start_props or {}),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **(end_props or {}),
    }


def generate_knows(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **(start_props or {}),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **(end_props or {}),
        "since": datetime.now().isoformat(),
    }


def generate_likes(
    id: str,
    start_id: str,
    start_type: str,
    start_props: dict,
    end_id: str,
    end_type: str,
    end_props: dict,
) -> dict:
    return {
        "id": id,
        "start_id": start_id,
        "start_vertex_type": start_type,
        **(start_props or {}),
        "end_id": end_id,
        "end_vertex_type": end_type,
        **(end_props or {}),
        "since": datetime.now().isoformat(),
    }


EDGE_GENERATORS: Dict[str, Any] = {
    "AirRoute": generate_airroute,  # For single generation fallback.
    "Bought": generate_bought,
    "Has": generate_has,
    "PerformedBy": generate_performedby,
    "UsedBy": generate_usedby,
    "UsedIn": generate_usedin,
    "Produce": generate_produce,
    "KNOWS": generate_knows,
    "LIKES": generate_likes,
}


# ------------------ CSV Writing ------------------
def sync_put_csv(data_dir: str, file_name: str, data: List[Dict[str, Any]]) -> str:
    if not data:
        return ""
    base_dir = os.path.abspath(data_dir)
    os.makedirs(base_dir, exist_ok=True)
    file_path = os.path.join(base_dir, f"{file_name.lower()}.csv")

    # Create CSV headers.
    headers = [f'"{k}"' for k in data[0].keys()]

    # Choose a buffer size (number of lines per flush)
    buffer_size = 100000

    # Open the file with an explicit buffering size (e.g., 1MB)
    with open(file_path, "w", encoding="utf-8", buffering=1_048_576) as f:
        # Write header
        f.write(",".join(headers) + "\n")

        # Use a list to accumulate lines before writing
        buffer_lines = []
        for i, row in enumerate(data, start=1):
            line = ",".join([f'"{v}"' for v in row.values()])
            buffer_lines.append(line)
            if i % buffer_size == 0:
                f.write("\n".join(buffer_lines) + "\n")
                buffer_lines = []
        # Write any remaining lines
        if buffer_lines:
            f.write("\n".join(buffer_lines))
    return file_path


async def put_csv(data_dir: str, file_name: str, data: List[Dict[str, Any]]) -> str:
    loop = asyncio.get_running_loop()
    result = await loop.run_in_executor(
        csv_executor, sync_put_csv, data_dir, file_name, data
    )
    return result


# ------------------ Multiprocessing Helpers ------------------
def generate_node_chunk(
    node_type: str, node_gen: Any, start_idx: int, end_idx: int
) -> List[Dict[str, Any]]:
    if node_type in BULK_SUPPORTED_NODE_TYPES:
        ids = [str(i + 1) for i in range(start_idx, end_idx)]
        return node_gen(ids)
    else:
        return [node_gen(str(i + 1)) for i in range(start_idx, end_idx)]


def generate_edge_chunk(
    start_idx: int,
    end_idx: int,
    edge_name: str,
    edge_props: Dict[str, Any],
    start_node_data: List[Dict[str, Any]],
    end_node_data: List[Dict[str, Any]],
) -> List[Dict[str, Any]]:
    # For AirRoute edges, use the bulk vectorized generator.
    if edge_name == "AirRoute":
        return generate_airroute_bulk(
            start_idx, end_idx, edge_props, start_node_data, end_node_data
        )
    else:
        gen = EDGE_GENERATORS[edge_name]
        chunk = []
        for i in range(start_idx, end_idx):
            start = random.choice(start_node_data)
            end = random.choice(end_node_data)
            edge = gen(
                str(i + 1),
                start["id"],
                edge_props["start"],
                start,
                end["id"],
                edge_props["end"],
                end,
            )
            chunk.append(edge)
        return chunk


# ------------------ Data Generation Functions ------------------
async def generate_complete_data(
    edge_name: str,
    edge_props: Dict[str, Any],
    nodes: Dict[str, int],
    data_dir: str,
) -> None:
    print(f"Generating {edge_name}: {edge_props['count']}...")
    start_node_gen = NODE_GENERATORS[edge_props["start"]]
    end_node_gen = NODE_GENERATORS[edge_props["end"]]
    if edge_props["start"] in BULK_SUPPORTED_NODE_TYPES:
        start_node_data = start_node_gen(
            [str(i + 1) for i in range(nodes[edge_props["start"]])]
        )
    else:
        start_node_data = [
            start_node_gen(str(i + 1)) for i in range(nodes[edge_props["start"]])
        ]

    if edge_props["end"] in BULK_SUPPORTED_NODE_TYPES:
        end_node_data = end_node_gen(
            [str(i + 1) for i in range(nodes[edge_props["end"]])]
        )
    else:
        end_node_data = [
            end_node_gen(str(i + 1)) for i in range(nodes[edge_props["end"]])
        ]
    total_edges = edge_props["count"]
    chunk_size = 100000  # Adjust as needed.
    tasks = []
    loop = asyncio.get_running_loop()
    with ProcessPoolExecutor(max_workers=os.cpu_count()) as executor:
        for start_idx in range(0, total_edges, chunk_size):
            end_idx = min(start_idx + chunk_size, total_edges)
            tasks.append(
                loop.run_in_executor(
                    executor,
                    generate_edge_chunk,
                    start_idx,
                    end_idx,
                    edge_name,
                    edge_props,
                    start_node_data,
                    end_node_data,
                )
            )
        results = await asyncio.gather(*tasks)
    data = [edge for sublist in results for edge in sublist]
    log.info(f"Writing {len(data)} edges.")
    await put_csv(
        data_dir=data_dir,
        file_name=f"{edge_props['start']}_{edge_props['end']}_{edge_name}",
        data=data,
    )


async def generate_nodes(
    cls_name: str, count: int, data_dir: str
) -> Tuple[str, List[Dict[str, Any]]]:
    print(f"Generating {cls_name}: {count}...")
    node_gen = NODE_GENERATORS[cls_name]
    chunk_size = 100000  # Adjust based on workload.
    tasks = []
    loop = asyncio.get_running_loop()
    with ProcessPoolExecutor(max_workers=os.cpu_count()) as executor:
        for start_idx in range(0, count, chunk_size):
            end_idx = min(start_idx + chunk_size, count)
            tasks.append(
                loop.run_in_executor(
                    executor,
                    generate_node_chunk,
                    cls_name,  # Pass the node type.
                    node_gen,
                    start_idx,
                    end_idx,
                )
            )
        results = await asyncio.gather(*tasks)
    data = [node for sublist in results for node in sublist]
    await put_csv(data_dir=data_dir, file_name=cls_name, data=data)
    return cls_name, data


async def generate_edges(
    edge_name: str,
    prop_list: List[Dict[str, Any]],
    nodes_data: Dict[str, List[Dict[str, Any]]],
    data_dir: str,
) -> None:
    loop = asyncio.get_running_loop()
    for props in prop_list:
        print(
            f"Generating {edge_name} {props['start']} - {props['end']}: {props['count']}..."
        )
        total_edges = props["count"]
        chunk_size = 100000  # Adjust based on workload.
        tasks = []
        start_node_data = nodes_data[props["start"]]
        end_node_data = nodes_data[props["end"]]
        with ProcessPoolExecutor(max_workers=os.cpu_count()) as executor:
            for start_idx in range(0, total_edges, chunk_size):
                end_idx = min(start_idx + chunk_size, total_edges)
                tasks.append(
                    loop.run_in_executor(
                        executor,
                        generate_edge_chunk,
                        start_idx,
                        end_idx,
                        edge_name,
                        props,
                        start_node_data,
                        end_node_data,
                    )
                )
            results = await asyncio.gather(*tasks)
        data = [edge for sublist in results for edge in sublist]
        await put_csv(
            data_dir=data_dir,
            file_name=f"{edge_name}_{props['start']}_{props['end']}",
            data=data,
        )


# ------------------ Main Function ------------------
async def main(
    pattern_no: int = 1, multiplier: int = 1, log_level: int = logging.INFO
) -> None:
    log.setLevel(log_level)
    if multiplier >= 100:
        log.warning("Multiplier is too large. It will generate a lot of data.")

    # Configuration constants
    SINGLE_SOURCE = 1
    MULTI_SOURCE = 2

    data_dir = "generated_dummy_" + datetime.now().strftime("%Y%m%d_%H%M%S")
    if not os.path.exists(data_dir):
        print(f"Creating directory {data_dir}")
        os.makedirs(data_dir)

    edges: Dict[str, List[Dict[str, Any]]] = {}
    sub_dir = ""
    nodes: Dict[str, int] = {}

    if pattern_no == 1:
        # Transaction: multiple node types, single edge type.
        source_type = SINGLE_SOURCE
        sub_dir = "/transaction"
        nodes = {"Customer": 10000 * multiplier, "Product": 1000 * multiplier}
        edges = {
            "Bought": [
                {"count": 20050 * multiplier, "start": "Customer", "end": "Product"}
            ]
        }
    elif pattern_no == 2:
        # Persons: single node type, multiple edge types.
        source_type = MULTI_SOURCE
        sub_dir = "/persons"
        nodes = {"Person": 1000 * multiplier}
        edges = {
            "KNOWS": [{"count": 1000 * multiplier, "start": "Person", "end": "Person"}],
            "LIKES": [{"count": 1000 * multiplier, "start": "Person", "end": "Person"}],
        }
    elif pattern_no == 3:
        # Airroute: single node and edge type.
        source_type = MULTI_SOURCE
        sub_dir = "/airroute"
        nodes = {"AirPort": 3500 * multiplier}
        edges = {
            "AirRoute": [
                {"count": 20000 * multiplier, "start": "AirPort", "end": "AirPort"}
            ]
        }
    elif pattern_no == 4:
        # Countries: multiple node types, single edge type.
        source_type = MULTI_SOURCE
        sub_dir = "/countries"
        nodes = {"Country": 200 * multiplier, "City": 10000 * multiplier}
        edges = {
            "Has": [{"count": 10000 * multiplier, "start": "Country", "end": "City"}]
        }
    elif pattern_no == 5:
        # Payment: multiple node types, multiple edge types.
        source_type = MULTI_SOURCE
        sub_dir = "/payment"
        nodes = {
            "BitcoinAddress": 900 * multiplier,
            "Cookie": 2700 * multiplier,
            "IP": 2200 * multiplier,
            "Phone": 960 * multiplier,
            "Email": 960 * multiplier,
            "Payment": 7000 * multiplier,
            "CreditCard": 1200 * multiplier,
            "PartnerEndUser": 400 * multiplier,
            "CryptoAddress": 160 * multiplier,
        }
        edges = {
            "UsedIn": [
                {"count": 6000 * multiplier, "start": "Cookie", "end": "Payment"},
                {"count": 6000 * multiplier, "start": "Email", "end": "Payment"},
                {
                    "count": 6000 * multiplier,
                    "start": "CryptoAddress",
                    "end": "Payment",
                },
                {"count": 6000 * multiplier, "start": "Phone", "end": "Payment"},
                {"count": 6000 * multiplier, "start": "CreditCard", "end": "Payment"},
            ],
            "PerformedBy": [
                {"count": 1000 * multiplier, "start": "Cookie", "end": "Payment"},
                {"count": 1000 * multiplier, "start": "Email", "end": "Payment"},
                {
                    "count": 1000 * multiplier,
                    "start": "CryptoAddress",
                    "end": "Payment",
                },
                {"count": 1000 * multiplier, "start": "Phone", "end": "Payment"},
                {"count": 1000 * multiplier, "start": "CreditCard", "end": "Payment"},
            ],
            "UsedBy": [
                {"count": 8000 * multiplier, "start": "Cookie", "end": "Payment"},
                {"count": 8000 * multiplier, "start": "Email", "end": "Payment"},
                {
                    "count": 8000 * multiplier,
                    "start": "CryptoAddress",
                    "end": "Payment",
                },
                {"count": 8000 * multiplier, "start": "Phone", "end": "Payment"},
                {"count": 8000 * multiplier, "start": "CreditCard", "end": "Payment"},
            ],
        }
    else:
        raise ValueError("Invalid TEST_PATTERN specified.")

    full_dir = f"{data_dir}{sub_dir}"
    if not os.path.exists(full_dir):
        print(f"Creating directory {full_dir}")
        os.makedirs(full_dir, exist_ok=True)

    if source_type == SINGLE_SOURCE:
        for edge_name, edge_props in edges.items():
            await generate_complete_data(
                edge_name=edge_name,
                edge_props=edge_props[0],
                nodes=nodes,
                data_dir=full_dir,
            )
    elif source_type == MULTI_SOURCE:
        node_tasks = [
            generate_nodes(cls_name, count, full_dir)
            for cls_name, count in nodes.items()
        ]
        nodes_list = await asyncio.gather(*node_tasks)
        nodes_data = {cls_name: data for cls_name, data in nodes_list}
        edge_tasks = [
            generate_edges(edge_name, prop_list, nodes_data, full_dir)
            for edge_name, prop_list in edges.items()
        ]
        await asyncio.gather(*edge_tasks)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import concurrent.futures
import logging
import sys
import time
from typing import Any, Dict

from agefreighter.csvdatamanager import CsvDataManager
from gremlin_python.driver import client, serializer  # type: ignore

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class CosmosGremlinLoader:
    """
    Load CSV data into Cosmos DB using the Gremlin API.
    """

    def __init__(
        self,
        csv_manager: CsvDataManager,
        cosmos_gremlin_endpoint: str,
        cosmos_key: str,
        cosmos_database: str,
        cosmos_container: str,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        self.csv_manager = csv_manager
        self.cosmos_gremlin_endpoint = cosmos_gremlin_endpoint
        self.cosmos_key = cosmos_key
        self.cosmos_database = cosmos_database
        self.cosmos_container = cosmos_container
        log.info("CosmosGremlinLoader initialized")

    def execute_gremlin_query(self, g_client: client.Client, query: str) -> None:
        """
        Execute a Gremlin query with retry logic.
        """
        retries = 0
        initial_wait = 1
        while retries < 5:
            try:
                future = g_client.submitAsync(query)
                result = future.result()
                log.debug(f"Gremlin query result: {result.all().result()}")
                return
            except Exception:
                wait_time = initial_wait * (2**retries)
                log.warning(
                    f"Query failed (attempt {retries + 1}). Retrying in {wait_time} seconds..."
                )
                time.sleep(wait_time)
                retries += 1
        raise Exception("Max retries exceeded for Gremlin query")

    async def load_data(self) -> None:
        log.info("Loading CSV to Cosmos DB via Gremlin API")

        start_time = time.time()
        COSMOS_USERNAME = "/dbs/db1/colls/transaction"
        COSMOS_PKEY = "pk"

        LOGICAL_PARTITION_SIZE = 20 * 1024 * 1024 * 1024  # 20GB
        AVERAGE_SIZE_OF_DOCUMENT = 512  # 512 bytes
        num_of_docs_per_partition = LOGICAL_PARTITION_SIZE // AVERAGE_SIZE_OF_DOCUMENT
        num_of_pk = 1
        MAX_OPERATOR_DEPTH = 400

        try:
            g_client = client.Client(
                url=self.cosmos_gremlin_endpoint,
                traversal_source="g",
                username=COSMOS_USERNAME,
                password=self.cosmos_key,
                message_serializer=serializer.GraphSONSerializersV2d0(),
                timeout=600,
            )
        except Exception as e:
            print(f"Failed to connect to Gremlin server: {e}")
            return

        df = self.csv_manager.get_dataframe()
        df.drop_duplicates(inplace=True)

        vertex_columns: Dict[str, Any] = {
            "Customer": ["CustomerID", "Name", "Address", "Email", "Phone"],
            "Product": [
                "ProductID",
                "Phrase",
                "SKU",
                "Price",
                "Color",
                "Size",
                "Weight",
            ],
        }

        total_docs = 0
        max_workers = 4
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for vertex_type, cols in vertex_columns.items():
                vertices = df[cols].drop_duplicates()
                # Escape single quotes in string fields
                vertices = vertices.applymap(
                    lambda x: x.replace("'", r"\'") if isinstance(x, str) else x
                )
                if vertex_type == "Customer":
                    tmp_query = """.addV('Customer')
                        .property('Name', '{Name}')
                        .property('CustomerID', '{CustomerID}')
                        .property('Address', '{Address}')
                        .property('Email', '{Email}')
                        .property('Phone', '{Phone}')
                        .property('{pk}', '{num_of_pk}')"""
                elif vertex_type == "Product":
                    tmp_query = """.addV('Product')
                        .property('Phrase', '{Phrase}')
                        .property('ProductID', '{ProductID}')
                        .property('SKU', '{SKU}')
                        .property('Price', '{Price}')
                        .property('Color', '{Color}')
                        .property('Size', '{Size}')
                        .property('Weight', '{Weight}')
                        .property('{pk}', '{num_of_pk}')"""
                chunk_size = int(MAX_OPERATOR_DEPTH / (len(cols) + 2))
                for i, chunk in enumerate(
                    CsvDataManager.get_chunks(vertices, chunk_size)
                ):
                    log.info(
                        f"Creating '{vertex_type}' vertices: {len(chunk)} records."
                    )
                    if len(chunk.columns) == 5:
                        query = "g" + "".join(
                            [
                                tmp_query.format(
                                    Name=row["Name"],
                                    CustomerID=row["CustomerID"],
                                    Address=row["Address"],
                                    Email=row["Email"],
                                    Phone=row["Phone"],
                                    pk=COSMOS_PKEY,
                                    num_of_pk=num_of_pk,
                                )
                                for _, row in chunk.iterrows()
                            ]
                        )
                    elif len(chunk.columns) == 7:
                        query = "g" + "".join(
                            [
                                tmp_query.format(
                                    Phrase=row["Phrase"],
                                    ProductID=row["ProductID"],
                                    SKU=row["SKU"],
                                    Price=row["Price"],
                                    Color=row["Color"],
                                    Size=row["Size"],
                                    Weight=row["Weight"],
                                    pk=COSMOS_PKEY,
                                    num_of_pk=num_of_pk,
                                )
                                for _, row in chunk.iterrows()
                            ]
                        )
                    futures.append(
                        executor.submit(self.execute_gremlin_query, g_client, query)
                    )
                    total_docs += len(chunk)
                    if total_docs % num_of_docs_per_partition == 0:
                        num_of_pk += 1
            concurrent.futures.wait(futures)

        # Create edges (BOUGHT relationships) using a larger thread pool
        max_workers = 1024
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = []
            for i, row in enumerate(df.itertuples(index=False), start=1):
                log.info(f"Creating 'BOUGHT' edge: {i}")
                edge_query = (
                    f"g.V().has('CustomerID', '{row.CustomerID}')"
                    f".addE('BOUGHT').to(g.V().has('ProductID', '{row.ProductID}'))"
                )
                futures.append(
                    executor.submit(self.execute_gremlin_query, g_client, edge_query)
                )
            concurrent.futures.wait(futures)

        g_client.close()
        self._show_time(start_time, sys._getframe().f_code.co_name)

    @staticmethod
    def _show_time(start_time: float, message: str) -> None:
        elapsed = time.time() - start_time if start_time else 0.0
        print(f"Time for {message}: {elapsed:.2f} seconds")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import sys
import time

from agefreighter.csvdatamanager import CsvDataManager
from neo4j import AsyncGraphDatabase

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class Neo4jLoader:
    """
    Load CSV data into Neo4j.
    """

    def __init__(
        self,
        csv_manager: CsvDataManager,
        neo4j_uri: str,
        neo4j_user: str,
        neo4j_password: str,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        self.csv_manager = csv_manager
        self.neo4j_uri = neo4j_uri
        self.neo4j_user = neo4j_user
        self.neo4j_password = neo4j_password
        log.info("Neo4jLoader initialized")

    async def load_data(self) -> None:
        log.info("Loading CSV to Neo4j")

        start_time = time.time()
        BATCH_SIZE = 1000
        df = self.csv_manager.get_dataframe()

        # Get unique start and end node information
        unique_starts = df[
            ["start_id", "start_vertex_type", "Name", "Address", "Email", "Phone"]
        ].drop_duplicates()
        unique_ends = df[
            ["end_id", "end_vertex_type", "SKU", "Price", "Color", "Size", "Weight"]
        ].drop_duplicates()

        start_label = unique_starts.iloc[0]["start_vertex_type"]
        end_label = unique_ends.iloc[0]["end_vertex_type"]

        async with AsyncGraphDatabase.driver(
            self.neo4j_uri, auth=(self.neo4j_user, self.neo4j_password)
        ) as driver:
            async with driver.session() as session:
                # Clear the database
                await session.run("MATCH (a)-[r]->() DELETE a, r")
                await session.run("MATCH (a) DELETE a")
                # Manage indices
                await session.run(f"DROP INDEX {start_label}_index_id IF EXISTS")
                await session.run(f"DROP INDEX {end_label}_index_id IF EXISTS")
                await session.run(
                    f"CREATE INDEX {start_label}_index_id FOR (n:{start_label}) ON (n.CustomerID)"
                )
                await session.run(
                    f"CREATE INDEX {end_label}_index_id FOR (n:{end_label}) ON (n.ProductID)"
                )

                # Create start nodes in batches
                for idx in range(0, len(unique_starts), BATCH_SIZE):
                    batch = unique_starts.iloc[idx : idx + BATCH_SIZE]
                    starts = [
                        {
                            start_label: row["start_vertex_type"],
                            "CustomerID": row["start_id"],
                            "Name": row["Name"],
                            "Address": row["Address"],
                            "Email": row["Email"],
                            "Phone": row["Phone"],
                        }
                        for _, row in batch.iterrows()
                    ]
                    query = (
                        f"UNWIND $starts AS row "
                        f"CREATE (a:{start_label} {{CustomerID: row.CustomerID, Name: row.Name, "
                        f"Address: row.Address, Email: row.Email, Phone: row.Phone}}) "
                        f"SET a += row"
                    )
                    await session.run(query, starts=starts)

                # Create end nodes in batches
                for idx in range(0, len(unique_ends), BATCH_SIZE):
                    batch = unique_ends.iloc[idx : idx + BATCH_SIZE]
                    ends = [
                        {
                            end_label: row["end_vertex_type"],
                            "ProductID": row["end_id"],
                            "SKU": row["SKU"],
                            "Price": row["Price"],
                            "Color": row["Color"],
                            "Size": row["Size"],
                            "Weight": row["Weight"],
                        }
                        for _, row in batch.iterrows()
                    ]
                    query = (
                        f"UNWIND $ends AS row "
                        f"CREATE (f:{end_label} {{ProductID: row.ProductID, SKU: row.SKU, "
                        f"Price: row.Price, Color: row.Color, Size: row.Size, Weight: row.Weight}}) "
                        f"SET f += row"
                    )
                    await session.run(query, ends=ends)

                # Create edges in batches
                for idx in range(0, len(df), BATCH_SIZE):
                    batch = df.iloc[idx : idx + BATCH_SIZE]
                    edges = [
                        {"from": row["start_id"], "to": row["end_id"]}
                        for _, row in batch.iterrows()
                    ]
                    query = (
                        f"UNWIND $edges AS row "
                        f"MATCH (from:{start_label} {{CustomerID: row.from}}) "
                        f"MATCH (to:{end_label} {{ProductID: row.to}}) "
                        f"CREATE (from)-[r:BOUGHT]->(to) "
                        f"SET r += row"
                    )
                    await session.run(query, edges=edges)
        self._show_time(start_time, sys._getframe().f_code.co_name)

    @staticmethod
    def _show_time(start_time: float, message: str) -> None:
        elapsed = time.time() - start_time if start_time else 0.0
        print(f"Time for {message}: {elapsed:.2f} seconds")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import sys
import time
from typing import List

from agefreighter.csvdatamanager import CsvDataManager
import pandas as pd
import psycopg as pg

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class PgsqlLoader:
    """
    Load CSV data into PostgreSQL.
    """

    def __init__(
        self, csv_manager: CsvDataManager, src_dsn: str, log_level: int = logging.INFO
    ) -> None:
        log.setLevel(log_level)
        self.csv_manager = csv_manager
        self.src_dsn = src_dsn

    async def load_data(self) -> None:
        log.info("Loading CSV to PGSQL")

        start_time = time.time()
        schema = "public"
        src_tables = {"start": "Customer", "end": "Product", "edges": "BOUGHT"}

        df = self.csv_manager.get_dataframe()

        # Prepare data and types for each table
        data_frames: List[pd.DataFrame] = [
            pd.DataFrame(),
            pd.DataFrame(),
            pd.DataFrame(),
        ]
        types: List[List[str]] = [[], [], []]

        # Start table (Customer)
        data_frames[0] = df[
            ["CustomerID", "Name", "Address", "Email", "Phone"]
        ].drop_duplicates()
        data_frames[0].insert(0, "CustomerSerial", range(1, len(data_frames[0]) + 1))
        types[0] = ["SERIAL", "TEXT", "TEXT", "TEXT", "TEXT", "TEXT"]

        # End table (Product)
        data_frames[1] = df[
            ["ProductID", "Phrase", "SKU", "Price", "Color", "Size", "Weight"]
        ].drop_duplicates()
        data_frames[1].insert(0, "ProductSerial", range(1, len(data_frames[1]) + 1))
        types[1] = ["SERIAL", "TEXT", "TEXT", "TEXT", "REAL", "TEXT", "TEXT", "INT"]

        # Edges table (BOUGHT)
        data_frames[2] = df[["CustomerID", "ProductID"]].copy()
        data_frames[2].insert(0, "BoughtSerial", range(1, len(data_frames[2]) + 1))
        types[2] = ["SERIAL", "TEXT", "TEXT"]

        with pg.connect(self.src_dsn) as conn:
            with conn.cursor() as cur:
                for (table_key, table_name), df_data, col_types in zip(
                    src_tables.items(), data_frames, types
                ):
                    cur.execute(f'DROP TABLE IF EXISTS {schema}."{table_name}"')
                    columns = ", ".join(
                        [f'"{col}" {tp}' for col, tp in zip(df_data.columns, col_types)]
                    )
                    cur.execute(f'CREATE TABLE {schema}."{table_name}" ({columns})')
                    query = (
                        f'COPY {schema}."{table_name}" FROM STDIN (FORMAT TEXT, FREEZE)'
                    )
                    with cur.copy(query) as copy:
                        copy_data = "\n".join(
                            "\t".join(map(str, row))
                            for row in df_data.itertuples(index=False)
                        )
                        copy.write(copy_data)
                    # Create indexes based on table type
                    if table_key == "edges":
                        cur.execute(
                            f'CREATE INDEX ON {schema}."{table_name}"("CustomerID")'
                        )
                        cur.execute(
                            f'CREATE INDEX ON {schema}."{table_name}"("ProductID")'
                        )
                    elif table_key == "start":
                        cur.execute(
                            f'CREATE INDEX ON {schema}."{table_name}"("CustomerID")'
                        )
                    elif table_key == "end":
                        cur.execute(
                            f'CREATE INDEX ON {schema}."{table_name}"("ProductID")'
                        )
                cur.execute("COMMIT")
        self._show_time(start_time, sys._getframe().f_code.co_name)

    @staticmethod
    def _show_time(start_time: float, message: str) -> None:
        elapsed = time.time() - start_time if start_time else 0.0
        print(f"Time for {message}: {elapsed:.2f} seconds")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for exporting PostgreSQL data for AGE COPY import using AgeFreighter.
"""

import asyncio
import json
import logging
import os
import sys
from typing import Any, Dict, List, Optional, cast, Tuple

from psycopg_pool import AsyncConnectionPool
from psycopg.sql import SQL, Identifier
from psycopg.rows import namedtuple_row

from .agefreighter import AgeFreighter

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class ConfigManager:
    """
    Loads and validates a JSON configuration file.
    """

    def __init__(
        self,
        config_path: str,
        src_con_pool: AsyncConnectionPool,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        self.config_path: str = config_path
        self.src_con_pool: AsyncConnectionPool = src_con_pool
        self.config_json: Dict[str, Any] = {}
        self.parse_result: Optional[str] = None

    def require_key(self, config: Dict[str, Any], key: str, context: str = "") -> Any:
        if key not in config:
            ctx = f" ({context})" if context else ""
            raise ValueError(
                f"Missing '{key}' key in config file {self.config_path}{ctx}"
            )
        return config[key]

    async def load_config(self) -> Dict[str, Any]:
        tables: List[str] = []

        # Load JSON configuration
        try:
            with open(os.path.abspath(self.config_path), "r", encoding="utf-8") as f:
                config_json = json.load(f)
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON format in config file {self.config_path}")

        edge_config = self.require_key(config_json, "edge", "root level")

        if isinstance(edge_config, dict):
            self.parse_result = "Config has single edge and multiple nodes"
            start_vertex = self.require_key(edge_config, "start_vertex", "edge config")
            end_vertex = self.require_key(edge_config, "end_vertex", "edge config")
            tables.extend(
                [
                    self.require_key(edge_config, "table", "edge config"),
                    self.require_key(start_vertex, "table", "start_vertex config"),
                    self.require_key(end_vertex, "table", "end_vertex config"),
                ]
            )
            for vertex in (start_vertex, end_vertex):
                self.require_key(vertex, "id", "vertex config")
                self.require_key(vertex, "label", "vertex config")
                self.require_key(vertex, "props", "vertex config")
            self.require_key(edge_config, "type", "edge config")
            self.require_key(edge_config, "props", "edge config")
            self.require_key(edge_config, "start_vertex", "edge config")
            self.require_key(edge_config, "end_vertex", "edge config")
        elif isinstance(edge_config, list):
            for ec in edge_config:
                self.require_key(ec, "type", "edge config")
                self.require_key(ec, "props", "edge config")
                if "vertex" in ec:
                    self.parse_result = "Config has multiple edges and single node"
                    tables.append(self.require_key(ec, "table", "edge config"))
                    vertex = self.require_key(ec, "vertex", "edge config")
                    self.require_key(vertex, "id", "vertex config")
                    self.require_key(vertex, "label", "vertex config")
                    self.require_key(vertex, "props", "vertex config")
                else:
                    self.parse_result = "Config has multiple edges and multiple nodes"
                    start_vertex = self.require_key(ec, "start_vertex", "edge config")
                    end_vertex = self.require_key(ec, "end_vertex", "edge config")
                    tables.extend(
                        [
                            self.require_key(ec, "table", "edge config"),
                            self.require_key(
                                start_vertex, "table", "start_vertex config"
                            ),
                            self.require_key(end_vertex, "table", "end_vertex config"),
                        ]
                    )
                    for vertex in (start_vertex, end_vertex):
                        self.require_key(vertex, "id", "vertex config")
                        self.require_key(vertex, "label", "vertex config")
                        self.require_key(vertex, "props", "vertex config")
        else:
            raise ValueError(f"Invalid configuration structure in {self.config_path}")

        if not tables:
            raise ValueError("No tables provided")

        async with self.src_con_pool.connection() as con:
            async with con.cursor() as cur:
                for table in tables:
                    await cur.execute(
                        "SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name = %s)",
                        (table,),
                    )
                    exists = await cur.fetchone()
                    if exists is None or not exists[0]:
                        raise ValueError(f"Table {table} does not exist")
        return config_json


class PGSQLExporter(AgeFreighter):
    """
    Exports nodes and edges from PostgreSQL tables into new CSV files for AGE COPY import.
    """

    def __init__(
        self,
        dsn: str,
        min_connections: int,
        max_connections: int,
        src_dsn: str,
        config: str,
        trial: bool,
        save_temps: bool,
        progress: bool,
        graph_name: str,
        chunk_size: int,
        log_level: int = logging.INFO,
    ) -> None:
        super().__init__(
            dsn=dsn,
            min_connections=min_connections,
            max_connections=max_connections,
            save_temps=save_temps,
            progress=progress,
            chunk_size=chunk_size,
            log_level=log_level,
        )
        log.setLevel(log_level)
        self.src_dsn = src_dsn
        self.config: str = config
        self.config_json: Dict[str, Any] = {}
        self.graph_name: str = graph_name
        self.trial: bool = trial
        self.trial_nodes_by_label: Dict[str, Dict[str, List[str]]] = {}
        self.id_maps: Dict[str, Dict[Any, int]] = {}
        self.no_of_edges_trial = 100
        self.log_level = log_level
        self.vertex_configs: Dict[str, Dict[str, Any]] = {}

    async def __aenter__(self) -> "PGSQLExporter":
        await super().__aenter__()
        log.debug("Entering async context for PGSQLExporter.")
        self.src_con_pool = await self.connect(
            dsn=self.src_dsn,
            max_connections=self.max_connections,
            min_connections=self.min_connections,
            max_attempts=self.max_attempts,
            retry_delay=self.retry_delay,
            **self.extra_kwargs,
        )
        config_manager = ConfigManager(
            config_path=self.config,
            src_con_pool=self.src_con_pool,
            log_level=self.log_level,
        )
        self.config_json = await config_manager.load_config()
        self._build_vertex_configs()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if exc_type:
            log.error("Exception in PGSQLFreighter context: %s", exc)
        if self.src_con_pool:
            try:
                await self.src_con_pool.close()
            except Exception as e:
                log.error("Error closing source connection pool: %s", e)
        log.debug("Exiting PGSQLFreighter context.")
        await super().__aexit__(exc_type, exc, tb)

    @staticmethod
    def _clean_row(row: Any) -> Dict[str, Any]:
        return {
            key.strip('"'): value.strip('"') if isinstance(value, str) else value
            for key, value in row._asdict().items()
        }

    def get_labels(self) -> List[str]:
        labels = set()
        edge_config = self.config_json["edge"]
        if isinstance(edge_config, dict):
            if "vertex" in edge_config:
                labels.add(edge_config["vertex"]["label"])
            else:
                labels.add(edge_config["start_vertex"]["label"])
                labels.add(edge_config["end_vertex"]["label"])
        elif isinstance(edge_config, list):
            for ec in edge_config:
                if "vertex" in ec:
                    labels.add(ec["vertex"]["label"])
                else:
                    labels.add(ec["start_vertex"]["label"])
                    labels.add(ec["end_vertex"]["label"])
        return list(labels)

    def get_ids(self) -> Dict[str, List[str]]:
        edge_config = self.config_json["edge"]
        if isinstance(edge_config, dict):
            return {
                "start_id": [edge_config["start_id"]],
                "end_id": [edge_config["end_id"]],
            }
        elif isinstance(edge_config, list):
            return {
                "start_id": [ec["start_id"] for ec in edge_config],
                "end_id": [ec["end_id"] for ec in edge_config],
            }
        return {}

    def get_relationship_types(self) -> List[str]:
        edge_config = self.config_json["edge"]
        types: List[str] = []
        if isinstance(edge_config, dict):
            edge_type = edge_config.get("type")
            if isinstance(edge_type, str):
                types.append(edge_type)
        elif isinstance(edge_config, list):
            for ec in edge_config:
                edge_type = ec.get("type")
                if isinstance(edge_type, str):
                    types.append(edge_type)
        return list(dict.fromkeys(types))

    def get_vertex_labels(self, rel_type: str) -> Tuple[str, Optional[str]]:
        edge_config = self.config_json["edge"]
        if isinstance(edge_config, dict):
            if edge_config.get("type") != rel_type:
                raise ValueError(
                    f"No vertex labels found for relationship type {rel_type}"
                )
            return (
                edge_config["start_vertex"]["label"],
                edge_config["end_vertex"]["label"],
            )
        elif isinstance(edge_config, list):
            for ec in edge_config:
                if ec.get("type") == rel_type:
                    if "vertex" in ec:
                        return (ec["vertex"]["label"], None)
                    return (ec["start_vertex"]["label"], ec["end_vertex"]["label"])
        raise ValueError(f"No vertex labels found for relationship type {rel_type}")

    async def _count_rows(self, table_name: str) -> int:
        assert self.src_con_pool is not None, (
            "Source connection pool is not initialized"
        )
        async with self.src_con_pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                await cur.execute(
                    SQL("SELECT COUNT(*) FROM {table_name}").format(
                        table_name=Identifier(table_name)
                    )
                )
                result = await cur.fetchone()
                if result is None:
                    return 0
                # Assuming result is a tuple, return its first element.
                return result[0]

    async def _fetch_edges_chunk_table(
        self, table_name: str, skip: int, chunk_size: int
    ) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        assert self.src_con_pool is not None, (
            "Source connection pool is not initialized"
        )
        async with self.src_con_pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                await cur.execute(
                    SQL(
                        "SELECT * FROM {table_name} LIMIT {limit} OFFSET {skip}"
                    ).format(
                        table_name=Identifier(table_name), limit=chunk_size, skip=skip
                    )
                )
                rows = await cur.fetchall()
                for idx, row in enumerate(rows):
                    cleaned = self._clean_row(row)
                    cleaned["_elementid"] = idx + 1  # temporary element id
                    results.append(cleaned)
        return results

    async def _fetch_nodes_chunk_table(
        self, table_name: str, skip: int, chunk_size: int, id_column: str
    ) -> List[Dict[str, Any]]:
        results: List[Dict[str, Any]] = []
        assert self.src_con_pool is not None, (
            "Source connection pool is not initialized"
        )
        async with self.src_con_pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                await cur.execute(
                    SQL(
                        "SELECT * FROM {table_name} LIMIT {limit} OFFSET {skip}"
                    ).format(
                        table_name=Identifier(table_name), limit=chunk_size, skip=skip
                    )
                )
                rows = await cur.fetchall()
                for row in rows:
                    cleaned = self._clean_row(row)
                    cleaned["_elementid"] = getattr(row, id_column)
                    results.append(cleaned)
        return results

    async def _fetch_nodes_by_ids_chunk_table(
        self, table_name: str, node_ids: List[str], id_column: str
    ) -> List[Dict[str, Any]]:
        # If no node IDs, return immediately to avoid syntax errors.
        if not node_ids:
            return []

        results: List[Dict[str, Any]] = []
        assert self.src_con_pool is not None, (
            "Source connection pool is not initialized"
        )
        async with self.src_con_pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                values = ",".join([str(nid) for nid in node_ids])
                await cur.execute(
                    'SELECT * FROM "{table_name}" WHERE {id_column} IN ({values})'.format(
                        table_name=table_name, id_column=id_column, values=values
                    )
                )
                rows = await cur.fetchall()
                for i, row in enumerate(rows):
                    cleaned = self._clean_row(row)
                    cleaned["_elementid"] = getattr(row, id_column)
                    results.append(cleaned)
        return results

    def get_edge_table_name(self, rel_type: str) -> str:
        edge_config = self.config_json["edge"]
        if isinstance(edge_config, dict):
            if edge_config.get("type") == rel_type:
                return edge_config["table"]
        elif isinstance(edge_config, list):
            for ec in edge_config:
                if ec.get("type") == rel_type:
                    return ec["table"]
        raise ValueError(f"No table found for relationship type {rel_type}")

    def _build_vertex_configs(self) -> None:
        vertex_configs: Dict[str, Any] = {}
        edge_config = self.config_json["edge"]

        if isinstance(edge_config, dict):
            # Single edge configuration (Pattern 1 and Pattern 2)
            if "start_vertex" in edge_config and "end_vertex" in edge_config:
                start_vc = edge_config["start_vertex"]
                end_vc = edge_config["end_vertex"]

                if "table" not in start_vc:
                    start_vc["table"] = edge_config["table"]
                if "table" not in end_vc:
                    end_vc["table"] = edge_config["table"]

                if start_vc["label"] == end_vc["label"]:
                    # Pattern 1: both vertices share the same label
                    vertex_configs[start_vc["label"]] = [start_vc, end_vc]
                else:
                    # Pattern 2: different labels, store as single dict per label
                    vertex_configs[start_vc["label"]] = start_vc
                    vertex_configs[end_vc["label"]] = end_vc
            elif "vertex" in edge_config:
                # In case there is a "vertex" key at the top level
                vc = edge_config["vertex"]
                if "table" not in vc:
                    vc["table"] = edge_config["table"]
                vertex_configs[vc["label"]] = vc
            else:
                raise ValueError("Invalid edge config structure")
        elif isinstance(edge_config, list):
            # Multiple edge configurations (Pattern 3 and Pattern 4)
            for ec in edge_config:
                if "vertex" in ec:
                    # Pattern 3: each edge has a "vertex" key
                    vc = ec["vertex"]
                    if "table" not in vc:
                        vc["table"] = ec["table"]
                    label = vc["label"]
                    if label not in vertex_configs:
                        vertex_configs[label] = vc
                    else:
                        existing = vertex_configs[label]
                        if isinstance(existing, list):
                            if vc not in existing:
                                existing.append(vc)
                        else:
                            if existing != vc:
                                vertex_configs[label] = [existing, vc]
                else:
                    # Pattern 4: each edge has "start_vertex" and "end_vertex"
                    for key in ["start_vertex", "end_vertex"]:
                        vc = ec[key]
                        if "table" not in vc:
                            vc["table"] = ec["table"]
                        label = vc["label"]
                        if label not in vertex_configs:
                            vertex_configs[label] = vc
                        else:
                            existing = vertex_configs[label]
                            if isinstance(existing, list):
                                if vc not in existing:
                                    existing.append(vc)
                            else:
                                if existing != vc:
                                    vertex_configs[label] = [existing, vc]
        else:
            raise ValueError("Invalid edge config type")

        self.vertex_configs = vertex_configs

    async def export_nodes(self) -> Dict[str, Dict[str, Any]]:
        """
        Export nodes from PostgreSQL to new CSV files and return mapping for AGE import.
        """
        vertex_args: Dict[str, Dict[str, Any]] = {}

        # Determine labels based on trial mode or full export
        if self.trial:
            # Wrap keys in list to ensure a list of strings.
            labels = list(list(self.trial_nodes_by_label.values())[0].keys())
        else:
            labels = list(self.vertex_configs.keys())

        for label in labels:
            try:
                await self.create_label_type(label_type="vertex", value=label)
                first_id = await self.get_first_id(self.graph_name, label)
                print(self.vertex_configs[label])
                if isinstance(self.vertex_configs[label], list):
                    vertex_config_list = cast(
                        List[Dict[str, Any]], self.vertex_configs[label]
                    )
                    tables = [vc["table"] for vc in vertex_config_list]
                    id_columns = [vc["id"] for vc in vertex_config_list]
                    properties = [vc["props"] for vc in vertex_config_list]
                else:
                    tables = [self.vertex_configs[label]["table"]]
                    id_columns = [self.vertex_configs[label]["id"]]
                    properties = [self.vertex_configs[label]["props"]]

                nodes: List[Dict[str, Any]] = []
                if self.trial:
                    for node_id_list in list(self.trial_nodes_by_label.values()):
                        node_ids = node_id_list[label]
                        for table, id_column in zip(tables, id_columns):
                            nodes.extend(
                                await self._fetch_nodes_by_ids_chunk_table(
                                    table, node_ids, id_column
                                )
                            )
                else:
                    for table, id_column in zip(tables, id_columns):
                        count = await self._count_rows(table)
                        tasks = [
                            self._fetch_nodes_chunk_table(
                                table, skip, int(self.chunk_size), id_column
                            )
                            for skip in range(0, count, int(self.chunk_size))
                        ]
                        for chunk in await asyncio.gather(*tasks):
                            nodes.extend(chunk)

                # Deduplicate nodes by their original ID.
                unique_nodes = {node["_elementid"]: node for node in nodes}
                nodes = list(unique_nodes.values())

                all_data = [
                    {
                        "id": idx + first_id,
                        "properties": {
                            k: v
                            for k, v in node.items()
                            if k in properties or k == "_elementid"
                        },
                    }
                    for idx, node in enumerate(nodes)
                ]

                self.id_maps[label] = {
                    cast(Dict[str, Any], item["properties"])["_elementid"]: cast(
                        int, item["id"]
                    )
                    for item in all_data
                    if "_elementid" in cast(Dict[str, Any], item["properties"])
                }

                file_path = await self.write_csv(label, "v", all_data)
                vertex_args[label] = {
                    "csv_path": file_path.replace("\\", "\\\\"),
                    "original_id": "_elementid",
                    "next_val": str(len(all_data)),
                }
            except Exception as exc:
                log.exception("Error exporting nodes for label '%s': %s", label, exc)
        return vertex_args

    async def export_edges(self) -> Dict[str, Dict[str, Any]]:
        """
        Export edges from PostgreSQL to new CSV files and return mapping for AGE import.
        """
        edge_args: Dict[str, Dict[str, Any]] = {}
        rel_types = self.get_relationship_types()

        for rel_type in rel_types:
            try:
                edge_table_name = self.get_edge_table_name(rel_type)
                await self.create_label_type(label_type="edge", value=rel_type)
                first_id = await self.get_first_id(self.graph_name, rel_type)
                count = await self._count_rows(edge_table_name)
                if self.trial:
                    count = min(count, 100)
                    limit = min(count, int(self.chunk_size))
                else:
                    limit = int(self.chunk_size)
                log.info(
                    "Exporting %d edges for relationship type '%s'.", count, rel_type
                )
                tasks = [
                    self._fetch_edges_chunk_table(edge_table_name, skip, limit)
                    for skip in range(0, count, limit)
                ]
                chunks = await asyncio.gather(*tasks)
                all_data: List[Dict[str, Any]] = []
                start_vertex_label, end_vertex_label = self.get_vertex_labels(rel_type)
                ids = self.get_ids()

                for sublist in chunks:
                    for idx, item in enumerate(sublist):
                        # Resolve start_id using candidate keys
                        new_start_id = None
                        for candidate in ids["start_id"]:
                            try:
                                new_start_id = self.id_maps[start_vertex_label][
                                    item[candidate]
                                ]
                                break
                            except KeyError:
                                continue
                        if new_start_id is None:
                            log.warning(
                                "Could not resolve start_id for record: %s", item
                            )
                            continue

                        # Resolve end_id using candidate keys
                        new_end_id = None
                        if end_vertex_label:
                            for candidate in ids["end_id"]:
                                try:
                                    new_end_id = self.id_maps[end_vertex_label][
                                        item[candidate]
                                    ]
                                    break
                                except KeyError:
                                    continue
                        else:
                            for candidate in ids["end_id"]:
                                try:
                                    new_end_id = self.id_maps[start_vertex_label][
                                        item[candidate]
                                    ]
                                    break
                                except KeyError:
                                    continue
                        if new_end_id is None:
                            log.warning("Could not resolve end_id for record: %s", item)
                            continue

                        edge_data = {
                            "id": idx + first_id,
                            "start_id": new_start_id,
                            "end_id": new_end_id,
                            "properties": {
                                k: v
                                for k, v in item.items()
                                if k
                                not in [
                                    "start_vertex_type",
                                    "start_id",
                                    "end_vertex_type",
                                    "end_id",
                                ]
                            },
                        }
                        all_data.append(edge_data)

                file_path = await self.write_csv(rel_type, "e", all_data)
                edge_args[rel_type] = {
                    "csv_path": file_path.replace("\\", "\\\\"),
                    "original_id": "_elementid",
                    "next_val": str(len(all_data)),
                }
            except Exception as exc:
                log.exception("Error exporting edges for '%s': %s", rel_type, exc)
        return edge_args

    async def list_nodes(self) -> None:
        """
        In trial mode, list nodes per relationship type by sampling edge rows.
        """
        for rel_type in self.get_relationship_types():
            try:
                edge_table_name = self.get_edge_table_name(rel_type)
                count = min(
                    await self._count_rows(edge_table_name),
                    self.no_of_edges_trial,
                )
                limit = min(int(self.chunk_size), count)
                log.info("Listing nodes for relationship type '%s'.", rel_type)
                tasks = [
                    self._fetch_edges_chunk_table(edge_table_name, skip, limit)
                    for skip in range(0, count, limit)
                ]
                chunks = await asyncio.gather(*tasks)
                all_data = [item for sublist in chunks for item in sublist]
                id_dict = self.get_ids()
                nodes_by_label: Dict[str, List[str]] = {}
                for node_label in self.vertex_configs.keys():
                    nodes_by_label[node_label] = []
                    for key_list in id_dict.values():
                        for candidate in key_list:
                            for record in all_data:
                                v = record.get(candidate, None)
                                if v:
                                    nodes_by_label[node_label].append(v)
                self.trial_nodes_by_label[rel_type] = nodes_by_label
            except Exception as exc:
                log.exception("Error listing nodes for '%s': %s", rel_type, exc)

    async def export(self) -> None:
        """
        Main export function to process both nodes and edges.
        """
        try:
            await self.set_up_graph(graph_name=self.graph_name, create_graph=True)
            if self.trial:
                await self.list_nodes()
            nodes_args = await self.export_nodes()
            if not nodes_args:
                log.error(
                    "No nodes exported.\nDoes the PostgreSQL contain data for nodes?"
                )
                sys.exit(1)
            edges_args = await self.export_edges()
            if not edges_args:
                log.error(
                    "No edges exported.\nDoes the PostgreSQL contain data for edges?"
                )
                sys.exit(1)
        except Exception as exc:
            log.exception("Error during export process: %s", exc)
            raise

        self.vertices = nodes_args
        self.edges = edges_args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
from datetime import datetime
import functools
import logging
import os
import platform
import sys
import tempfile
from typing import Dict, Any, List, Set, Optional

import numpy as np
from psycopg_pool import AsyncConnectionPool, PoolTimeout
from psycopg.sql import SQL, Identifier, Literal
from psycopg.rows import namedtuple_row
import aiofiles

# Configure logging; default to INFO (overridable via --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)

# Console formatting constants
RED = "\033[91m"
BOLD = "\033[1m"
RESET = "\033[0m"


def reconnect_on_failure(func):
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        attempts = 0
        while attempts < self.max_attempts:
            try:
                return await func(self, *args, **kwargs)
            except ValueError:
                # Immediately re-raise ValueError (non-transient)
                raise
            except Exception as e:
                attempts += 1
                log.error("Error in %s: %s (attempt %d)", func.__name__, e, attempts)
                # (Connection-recovery logic omitted for brevity)
                await asyncio.sleep(self.retry_delay)
        log.error("Max attempts reached in %s", func.__name__)
        raise Exception(f"Max attempts reached in {func.__name__}")

    return wrapper


class AgeFreighter:
    """
    Manages a PostgreSQL connection pool, sets up an AGE graph,
    and copies CSV data using PostgreSQL's COPY protocol.
    """

    def __init__(
        self,
        dsn: str = "",
        max_connections: int = 64,
        min_connections: int = 4,
        save_temps: bool = False,
        progress: bool = True,
        output_dir: Optional[str] = None,
        chunk_size: int = 8192,
        graph_name: str = "",
        log_level: int = logging.INFO,
        **kwargs,
    ) -> None:
        log.setLevel(log_level)
        log.debug("Initializing AgeFreighter.")
        self.dsn = dsn
        self.max_connections = max_connections
        self.min_connections = min_connections
        self.chunk_size = chunk_size
        self.graph_name = graph_name
        self.save_temps = save_temps
        self.progress = progress
        self.output_dir = output_dir
        self.con_pool: Optional[AsyncConnectionPool] = None
        self.vertices: Dict[str, Any] = {}
        self.edges: Dict[str, Any] = {}
        self.extra_kwargs: Dict[str, Any] = kwargs
        self.max_attempts = 3
        self.retry_delay = 3

        # Append search_path options for AGE
        self.dsn_w_option = (
            f"{dsn} options='-c search_path=ag_catalog,\"$user\",public'"
        )

        if self.save_temps:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            self.output_dir = os.path.join(os.getcwd(), f"exported_{timestamp}")
            if not os.path.exists(self.output_dir):
                log.info(
                    "Output directory '%s' does not exist. Creating it.",
                    self.output_dir,
                )
                try:
                    os.makedirs(self.output_dir)
                except Exception as e:
                    log.error(
                        "Error creating output directory '%s': %s", self.output_dir, e
                    )
                    raise

    def __del__(self):
        # Cleanup temporary CSV files if not preserving temps.
        if not self.save_temps:
            try:
                combined = {**self.vertices, **self.edges}
                for csv_spec in combined.values():
                    csv_path = csv_spec.get("csv_path")
                    if csv_path and os.path.exists(csv_path):
                        log.info("Deleting CSV file: %s", csv_path)
                        os.remove(csv_path)
            except Exception as e:
                log.error("Error during cleanup of temporary files: %s", e)

    async def __aenter__(self) -> "AgeFreighter":
        log.debug("Entering async context for AgeFreighter.")
        self.con_pool = await self.connect(
            dsn=self.dsn_w_option,
            max_connections=self.max_connections,
            min_connections=self.min_connections,
            max_attempts=self.max_attempts,
            retry_delay=self.retry_delay,
            **self.extra_kwargs,
        )
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if exc_type:
            log.error("Exception in AgeFreighter context: %s", exc)
        if self.con_pool:
            try:
                await self.con_pool.close()
            except Exception as e:
                log.error("Error closing connection pool: %s", e)
        log.debug("Exiting AgeFreighter context.")

    async def connect(
        self,
        dsn: str,
        max_connections: int,
        min_connections: int,
        max_attempts: int,
        retry_delay: int,
        **kwargs,
    ) -> AsyncConnectionPool:
        """
        Open a connection pool with retry logic.
        """
        parsed_dsn = self.parse_dsn(dsn)
        log.debug("Opening connection pool for %s.", parsed_dsn["host"])

        # Increase file descriptor limits for Unix systems if needed
        if platform.system() in ["Darwin", "Linux"]:
            try:
                import resource

                current_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
                resource.setrlimit(resource.RLIMIT_NOFILE, (8192, current_limit[1]))
            except Exception as e:
                log.error("Error setting resource limits: %s", e)

        con_pool = AsyncConnectionPool(
            dsn,
            max_size=max_connections,
            min_size=min_connections,
            open=False,
            timeout=600,
            **kwargs,
        )

        for attempt in range(1, max_attempts + 1):
            try:
                await con_pool.open()
                await con_pool.wait()
                log.info(
                    "Connection pool for %s opened successfully.", parsed_dsn["host"]
                )
                break
            except PoolTimeout as e:
                log.error(
                    "Connection pool for %s timeout on attempt %d: %s",
                    parsed_dsn["host"],
                    attempt,
                    e,
                )
                if attempt == max_attempts:
                    raise
                await asyncio.sleep(retry_delay)
            except Exception as e:
                log.error(
                    "Error opening connection pool for %s on attempt %d: %s",
                    parsed_dsn["host"],
                    attempt,
                    e,
                )
                raise

        return con_pool

    async def close(self) -> None:
        """
        Close the connection pool.
        """
        log.debug("Closing connection pool.")
        if self.con_pool:
            try:
                await self.con_pool.close()
            except Exception as e:
                log.error("Error closing connection pool: %s", e)

    @staticmethod
    def parse_dsn(dsn: str) -> dict:
        """
        Parse a DSN string into an AST.
        """
        try:
            return dict(item.split("=", 1) for item in dsn.split())
        except ValueError:
            raise ValueError("Invalid DSN format")

    @reconnect_on_failure
    async def _recover_label(self, label: str) -> None:
        """
        Recover the label/type relation.
        """
        log.info("Recovering label/type relation '%s'.", label)
        pool = self.con_pool
        assert pool is not None, "Connection pool is not initialized."
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                log.info("Truncating label '%s'.", label)
                await cur.execute(
                    SQL("TRUNCATE {graph_name}.{label};").format(
                        graph_name=Identifier(self.graph_name),
                        label=Identifier(label),
                    )
                )
                log.info("Resetting sequence for label '%s'.", label)
                await cur.execute(
                    f'SELECT setval(\'"{self.graph_name}"."{label}_id_seq"\', 1, false)'
                )

    async def _retry_copy(
        self, label: str, csv_spec: Dict[str, Any], kind: str
    ) -> None:
        """
        Helper to retry a COPY operation for a given label.
        """
        attempts = 0
        while attempts < self.max_attempts:
            try:
                attempts += 1
                await self._copy(
                    graph_name=self.graph_name,
                    csv_path=csv_spec["csv_path"],
                    label_name=label,
                    next_val=csv_spec["next_val"],
                    kind=kind,
                )
                return
            except Exception:
                log.exception("Failed to copy '%s' (attempt %d)", label, attempts)
                if attempts >= self.max_attempts:
                    log.error("Max attempts reached for copying '%s'", label)
                    sys.exit(1)
                await self._recover_label(label=label)
                await asyncio.sleep(self.retry_delay)

    @reconnect_on_failure
    async def _copy_vertices(self) -> None:
        """
        Copy all vertex CSV files into the AGE graph.
        """
        for vertex_label, csv_spec in self.vertices.items():
            log.info("Copying vertex '%s' using CSV spec: %s", vertex_label, csv_spec)
            await self._retry_copy(vertex_label, csv_spec, kind="v")

    @reconnect_on_failure
    async def _copy_edges(self) -> None:
        """
        Copy all edge CSV files into the AGE graph.
        """
        for edge_type, csv_spec in self.edges.items():
            log.info("Copying edge '%s' using CSV spec: %s", edge_type, csv_spec)
            await self._retry_copy(edge_type, csv_spec, kind="e")

    @reconnect_on_failure
    async def _copy(
        self,
        graph_name: str,
        csv_path: str,
        label_name: str,
        next_val: int,
        kind: str,
    ) -> None:
        """
        Execute the COPY command to load CSV data.
        """
        log.debug("Starting COPY for file '%s' into label '%s'.", csv_path, label_name)
        # graph_name_quoted = self.quoted_graph_name(graph_name)
        if kind == "v":
            query = SQL(
                "COPY {graph_name}.{label_name} FROM STDIN (FORMAT CSV)"
            ).format(
                graph_name=Identifier(graph_name), label_name=Identifier(label_name)
            )
        else:
            query = SQL(
                "COPY {graph_name}.{label_name} (id, start_id, end_id, properties) FROM STDIN (FORMAT CSV)"
            ).format(
                graph_name=Identifier(graph_name), label_name=Identifier(label_name)
            )

        pool = self.con_pool
        assert pool is not None, "Connection pool is not initialized."
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                try:
                    async with cur.copy(query) as copy:
                        async with aiofiles.open(
                            csv_path, "r", encoding="utf-8"
                        ) as file:
                            while data := await file.read(self.chunk_size):
                                await copy.write(data)
                    log.info("Finished COPY for '%s'.", csv_path)
                except Exception as e:
                    log.error("Error during COPY from file '%s': %s", csv_path, e)
                    raise
                # Reset sequence after COPY
                await cur.execute(
                    f'SELECT setval(\'"{self.graph_name}"."{label_name}_id_seq"\', {next_val}, true)'
                )

                await cur.execute(SQL("COMMIT"))

    @staticmethod
    def extract_unique_keys(data: List[Dict[str, Any]]) -> Set[str]:
        """
        Extract unique property keys from a list of records.
        """
        unique_keys = set()
        for row in data:
            unique_keys.update(row.get("properties", {}).keys())
        return unique_keys

    @staticmethod
    def quoted_graph_name(graph_name: str) -> str:
        """
        Return the properly quoted graph name.
        """
        log.debug("Quoting graph name: %s", graph_name)
        return f'"{graph_name}"' if graph_name.lower() != graph_name else graph_name

    async def copy(self) -> None:
        """
        Copy vertices and edges from CSV files into the AGE graph.
        """
        if not self.graph_name:
            raise ValueError("graph_name must be specified")
        if not self.vertices:
            raise ValueError("vertices must be specified")
        if not self.edges:
            raise ValueError("edges must be specified")

        await self._copy_vertices()
        await self._copy_edges()

    @reconnect_on_failure
    async def get_first_id(self, graph_name: str, label_type: str) -> int:
        """
        Get the first ID for a given label type.
        """
        quoted_graph = self.quoted_graph_name(graph_name)
        pool = self.con_pool
        assert pool is not None, "Connection pool is not initialized."
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                relation = f'{quoted_graph}."{label_type}"'
                await cur.execute(
                    SQL(
                        "SELECT id FROM ag_label WHERE relation = {}::regclass;"
                    ).format(Literal(relation))
                )
                row = await cur.fetchone()
                if row is None:
                    raise ValueError("No row returned; check your query or data.")
                ENTRY_ID_BITS = 32 + 16
                ENTRY_ID_MASK = np.uint64(0x0000FFFFFFFFFFFF)
                first_id = ((np.uint64(row[0])) << ENTRY_ID_BITS) | (
                    np.uint64(1) & ENTRY_ID_MASK
                )
                return int(first_id)

    @reconnect_on_failure
    async def set_up_graph(self, graph_name: str, create_graph: bool = False) -> None:
        """
        Create or validate the graph in the PostgreSQL database.
        """
        log.debug("Setting up graph '%s'.", graph_name)
        # Remove surrounding quotes if present.
        self.graph_name = Identifier(graph_name).as_string().strip('"')
        pool = self.con_pool
        assert pool is not None, "Connection pool is not initialized."
        async with pool.connection() as conn:
            async with conn.cursor(row_factory=namedtuple_row) as cur:
                try:
                    await cur.execute(SQL("CREATE EXTENSION IF NOT EXISTS age CASCADE"))
                    await cur.execute(
                        SQL("SELECT count(*) FROM ag_graph WHERE name = {name}").format(
                            name=Literal(self.graph_name)
                        )
                    )
                    row = await cur.fetchone()
                    exists = row is not None and row.count == 1
                    if exists:
                        log.debug("Graph '%s' already exists.", self.graph_name)
                        if create_graph:
                            log.debug(
                                "Dropping and recreating graph '%s'.", self.graph_name
                            )
                            await cur.execute(
                                SQL("SELECT drop_graph({name}, true)").format(
                                    name=Literal(self.graph_name)
                                )
                            )
                            await cur.execute(
                                SQL("SELECT create_graph({name})").format(
                                    name=Literal(self.graph_name)
                                )
                            )
                        else:
                            log.debug("Using existing graph '%s'.", self.graph_name)
                    else:
                        log.debug("Graph '%s' does not exist.", self.graph_name)
                        if create_graph:
                            log.debug("Creating graph '%s'.", self.graph_name)
                            await cur.execute(
                                SQL("SELECT create_graph({name})").format(
                                    name=Literal(self.graph_name)
                                )
                            )
                        else:
                            raise ValueError(
                                f"Graph '{self.graph_name}' doesn't exist. Set create_graph=True."
                            )
                except Exception as e:
                    log.error("Error setting up graph '%s': %s", self.graph_name, e)
                    raise

    @reconnect_on_failure
    async def create_label_type(self, label_type: str, value: str) -> None:
        """
        Create a vertex or edge label type in the AGE graph.
        """
        log.debug("Creating %s label type with value '%s'.", label_type, value)
        pool = self.con_pool
        assert pool is not None, "Connection pool is not initialized."
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                try:
                    if label_type == "vertex":
                        log.debug("Creating vertex label '%s'.", value)
                        await cur.execute(
                            SQL("SELECT create_vlabel({schema}, {label});").format(
                                schema=Literal(self.graph_name),
                                label=Literal(value),
                            )
                        )
                        log.debug("Creating indices for vertex '%s'.", value)
                        await cur.execute(
                            SQL(
                                "CREATE INDEX ON {schema}.{label} USING GIN (properties);"
                            ).format(
                                schema=Identifier(self.graph_name),
                                label=Identifier(value),
                            )
                        )
                        await cur.execute(
                            SQL(
                                "CREATE INDEX ON {schema}.{label} USING BTREE (id);"
                            ).format(
                                schema=Identifier(self.graph_name),
                                label=Identifier(value),
                            )
                        )
                    elif label_type == "edge":
                        log.debug("Creating edge label '%s'.", value)
                        await cur.execute(
                            SQL("SELECT create_elabel({schema}, {label});").format(
                                schema=Literal(self.graph_name),
                                label=Literal(value),
                            )
                        )
                        log.debug("Creating indices for edge '%s'.", value)
                        await cur.execute(
                            SQL("CREATE INDEX ON {schema}.{label} (start_id);").format(
                                schema=Identifier(self.graph_name),
                                label=Identifier(value),
                            )
                        )
                        await cur.execute(
                            SQL("CREATE INDEX ON {schema}.{label} (end_id);").format(
                                schema=Identifier(self.graph_name),
                                label=Identifier(value),
                            )
                        )
                    else:
                        raise ValueError(f"Unsupported label type: {label_type}")
                except Exception as e:
                    log.error(
                        "Error creating label type '%s' with value '%s': %s",
                        label_type,
                        value,
                        e,
                    )
                    raise

    async def write_csv(self, label: str, kind: str, data: List[Dict[str, Any]]) -> str:
        """
        Write exported data to a CSV file.
        For nodes (kind "v") and edges (kind "e") different formats are used.
        If any row contains a tab, it is also written to an extra CSV file.

        Returns:
            The main CSV file path.
        """
        if not data:
            log.info("No data to write for '%s'.", label)
            return ""

        def format_kv(key: str, value: Any) -> str:
            safe_value = str(value).replace("\t", "\\t").replace('"', '\\""')
            return f'""{key}"": ""{safe_value}""'

        normal_file_path = (
            os.path.join(self.output_dir, f"{label.lower()}.csv")
            if self.output_dir
            else tempfile.NamedTemporaryFile(delete=False).name
        )

        tab_file_path: Optional[str] = None
        tab_file = None
        tab_printed = False
        headers = self.extract_unique_keys(data)
        BATCH_SIZE = 10000

        try:
            async with aiofiles.open(
                normal_file_path, "w", encoding="utf-8", newline=""
            ) as normal_f:
                normal_lines: List[str] = []
                tab_lines: List[str] = []
                for row in data:
                    props = row.get("properties", {})
                    formatted_parts = [
                        format_kv(h, props.get(h, ""))
                        for h in headers
                        if props.get(h, "")
                    ]
                    line = ", ".join(formatted_parts)
                    if kind == "e":
                        csv_line = f'{row["id"]},{row["start_id"]},{row["end_id"]},"{{{line}}}"\n'
                    elif kind == "v":
                        csv_line = f'{row["id"]},"{{{line}}}"\n'
                    else:
                        raise ValueError(f"Unsupported kind: {kind}")
                    normal_lines.append(csv_line)
                    if any("\t" in str(props.get(h, "")) for h in headers):
                        tab_lines.append(csv_line)
                    if len(normal_lines) >= BATCH_SIZE:
                        await normal_f.write("".join(normal_lines))
                        normal_lines = []
                    if tab_lines and len(tab_lines) >= BATCH_SIZE:
                        if tab_file is None:
                            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                            tab_dir = os.path.join(
                                os.getcwd(), f"tab_replaced_{timestamp}"
                            )
                            os.makedirs(tab_dir, exist_ok=True)
                            tab_file_path = os.path.join(
                                tab_dir, f"{label.lower()}_tab_replaced.csv"
                            )
                            tab_file = await aiofiles.open(
                                tab_file_path, "w", encoding="utf-8", newline=""
                            )
                        await tab_file.write("".join(tab_lines))
                        tab_lines = []
                if normal_lines:
                    await normal_f.write("".join(normal_lines))
                if tab_lines:
                    if tab_file is None:
                        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                        tab_dir = os.path.join(os.getcwd(), f"tab_replaced_{timestamp}")
                        os.makedirs(tab_dir, exist_ok=True)
                        tab_file_path = os.path.join(
                            tab_dir, f"{label.lower()}_tab_replaced.csv"
                        )
                        tab_file = await aiofiles.open(
                            tab_file_path, "w", encoding="utf-8", newline=""
                        )
                    await tab_file.write("".join(tab_lines))
            log.info("Exported %d records to %s.", len(data), normal_file_path)
        except Exception as e:
            log.exception("Error writing CSV for '%s': %s", label, e)
            raise
        finally:
            if tab_file is not None:
                await tab_file.close()
                if not tab_printed and tab_file_path:
                    print(
                        f"{RED}{BOLD}Tab replacement occurred. Additional rows have been output to: {tab_file_path}{RESET}"
                    )
                    tab_printed = True

        return normal_file_path.replace("\\", "\\\\")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module for exporting data from a Neo4j database to CSV files and loading them using AgeFreighter.
"""

import asyncio
import concurrent.futures
import logging
import sys
from typing import List, Dict, Any, Union

from neo4j import GraphDatabase, Result  # type: ignore
from .agefreighter import AgeFreighter

# Configure logging; default to INFO (overridable by the --debug flag)
logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


class Neo4jExporter(AgeFreighter):
    """
    Exports nodes and edges from a Neo4j database into CSV files.
    """

    def __init__(
        self,
        dsn: str,
        min_connections: int,
        max_connections: int,
        uri: str,
        user: str,
        password: str,
        database: str,
        trial: bool,
        save_temps: bool,
        progress: bool,
        graph_name: str,
        chunk_size: int,
        log_level: int = logging.INFO,
    ) -> None:
        log.setLevel(log_level)
        super().__init__(
            dsn=dsn,
            min_connections=min_connections,
            max_connections=max_connections,
            save_temps=save_temps,
            progress=progress,
            chunk_size=chunk_size,
            log_level=log_level,
        )
        self.uri: str = uri
        self.user: str = user
        self.password: str = password
        self.database: str = database  # optional
        self.graph_name: str = graph_name
        self.trial: bool = trial
        self.trial_nodes_by_label: Dict[str, Dict[str, List[str]]] = {}
        self.id_maps: Dict[str, Dict[str, int]] = {}
        self.max_attempts = 3
        self.retry_delay = 1
        self.no_of_edges_trial = 100

        try:
            self.driver = GraphDatabase.driver(uri, auth=(user, password))
            with self.driver.session() as session:
                session.run("RETURN 1").single()
            log.info("Connected to Neo4j at %s.", uri)
        except Exception as e:
            log.error("Error connecting to Neo4j: %s", e)
            sys.exit(1)

    async def __aenter__(self) -> "Neo4jExporter":
        await super().__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        try:
            self.driver.close()
            log.debug("Closed Neo4j driver.")
        except Exception as e:
            log.error("Error closing Neo4j driver: %s", e)
        await super().__aexit__(exc_type, exc_value, traceback)

    def get_labels(self) -> List[str]:
        """
        Retrieve all node labels from Neo4j.
        """
        try:
            with self.driver.session() as session:
                query = "CALL db.labels() YIELD label RETURN label;"
                result: Result = session.run(query)
                labels = [record["label"] for record in result]
                # Check for nodes with no labels
                query_no_label = (
                    "MATCH (n) WHERE size(labels(n)) = 0 RETURN COUNT(n) AS cnt"
                )
                result_no_label: Result = session.run(query_no_label)
                record = result_no_label.single()
                if record and int(record["cnt"]) > 0:
                    labels.append("NO_LABEL")
            return labels
        except Exception as e:
            log.exception("Error retrieving labels: %s", e)
            return []

    def get_relationship_types(self) -> List[str]:
        """
        Retrieve all relationship types from Neo4j.
        """
        try:
            with self.driver.session() as session:
                query = "CALL db.relationshipTypes() YIELD relationshipType RETURN relationshipType;"
                result: Result = session.run(query)
                types = [record["relationshipType"] for record in result]
            return types
        except Exception as e:
            log.exception("Error retrieving relationship types: %s", e)
            return []

    def _count_nodes(self, label: Union[str, List[str]]) -> int:
        """
        Count the number of nodes for a given label.
        """
        try:
            with self.driver.session() as session:
                if isinstance(label, str):
                    if label == "NO_LABEL":
                        query = (
                            "MATCH (n) WHERE size(labels(n)) = 0 RETURN COUNT(n) AS cnt"
                        )
                    else:
                        query = f"MATCH (n:{label}) RETURN COUNT(n) AS cnt"
                else:
                    labels_str = ", ".join(f'"{lbl}"' for lbl in label)
                    query = f"MATCH (n) WHERE ANY(lbl IN labels(n) WHERE lbl IN [{labels_str}]) RETURN COUNT(n) AS cnt"
                result: Result = session.run(query)
                record = result.single()
                return int(record["cnt"]) if record else 0
        except Exception as e:
            log.exception("Error counting nodes for label %s: %s", label, e)
            return 0

    def _count_edges(self, rel_type: str) -> int:
        """
        Count the number of edges for a given relationship type.
        """
        try:
            with self.driver.session() as session:
                query = f"MATCH ()-[r:{rel_type}]->() RETURN COUNT(r) AS cnt"
                result: Result = session.run(query)
                record = result.single()
                return int(record["cnt"]) if record else 0
        except Exception as e:
            log.exception("Error counting edges for '%s': %s", rel_type, e)
            return 0

    def _fetch_nodes_chunk(self, query: str) -> List[Dict[str, Any]]:
        """
        Fetch a chunk of nodes based on the provided Cypher query.
        """
        try:
            with self.driver.session() as session:
                result = session.run(query)
                data = [
                    {"_elementid": record["n"].element_id, **record["n"]._properties}
                    for record in result
                ]
            return data
        except Exception as e:
            log.exception("Error fetching nodes chunk: %s", e)
            return []

    def _fetch_nodes_by_ids_chunk(
        self, label: str, node_ids: List[str]
    ) -> List[Dict[str, Any]]:
        """
        Fetch nodes for a given label with specific element IDs.
        """
        try:
            with self.driver.session() as session:
                if label == "NO_LABEL":
                    query = "MATCH (n) WHERE size(labels(n)) = 0 AND elementId(n) IN $node_ids RETURN n"
                else:
                    query = (
                        f"MATCH (n:{label}) WHERE elementId(n) IN $node_ids RETURN n"
                    )
                result = session.run(query, node_ids=node_ids)
                data = [
                    {"_elementid": record["n"].element_id, **record["n"]._properties}
                    for record in result
                ]
            return data
        except Exception as e:
            log.exception("Error fetching nodes by IDs for label '%s': %s", label, e)
            return []

    def _fetch_edge_chunk(
        self, rel_type: str, skip: int, limit: int
    ) -> List[Dict[str, Any]]:
        """
        Fetch a chunk of edges for a given relationship type using pagination.
        """
        try:
            with self.driver.session() as session:
                query = f"MATCH (m)-[r:{rel_type}]->(n) SKIP {skip} LIMIT {limit} RETURN m, r, n"
                result = session.run(query)
                data = []
                for record in result:
                    m_labels = (
                        list(record["m"].labels) if record["m"].labels else ["NO_LABEL"]
                    )
                    n_labels = (
                        list(record["n"].labels) if record["n"].labels else ["NO_LABEL"]
                    )
                    for m_label in m_labels:
                        for n_label in n_labels:
                            entry = {
                                "_elementid": record["r"].element_id,
                                "start_id": record["m"].element_id,
                                "start_vertex_type": m_label,
                                "end_id": record["n"].element_id,
                                "end_vertex_type": n_label,
                                **record["r"]._properties,
                            }
                            data.append(entry)
            return data
        except Exception as e:
            log.exception(
                "Error fetching edge chunk for '%s' (skip %d): %s", rel_type, skip, e
            )
            return []

    async def export_nodes(
        self, thread_pool: concurrent.futures.ThreadPoolExecutor
    ) -> Dict[str, Dict[str, Any]]:
        """
        Export nodes from Neo4j to CSV files and return a mapping for AGE import.
        """
        loop = asyncio.get_running_loop()
        vertex_args: Dict[str, Dict[str, Any]] = {}
        combined_nodes: Dict[str, List[str]] = {}

        # In trial mode, combine provided node ID mappings; otherwise, fetch all labels.
        if self.trial:
            for mapping in self.trial_nodes_by_label.values():
                for label, ids in mapping.items():
                    combined_nodes.setdefault(label, []).extend(ids)
            labels = list(combined_nodes.keys())
        else:
            labels = self.get_labels()

        for label in labels:
            try:
                # Create the vertex label in AGE
                await self.create_label_type(label_type="vertex", value=label)
            except Exception as e:
                log.error("Error creating label '%s': %s", label, e)
                sys.exit(1)

            try:
                first_id = await self.get_first_id(self.graph_name, label)
            except Exception as e:
                log.error("Error getting first ID for label '%s': %s", label, e)
                sys.exit(1)

            count = (
                len(combined_nodes.get(label, []))
                if self.trial
                else self._count_nodes(label)
            )
            log.info("Exporting %d nodes for label '%s'.", count, label)
            try:
                if self.trial:
                    tasks = [
                        loop.run_in_executor(
                            thread_pool,
                            self._fetch_nodes_by_ids_chunk,
                            label,
                            combined_nodes[label][skip : skip + int(self.chunk_size)],
                        )
                        for skip in range(0, count, int(self.chunk_size))
                    ]
                else:
                    query_template = (
                        "MATCH (n) WHERE size(labels(n)) = 0 SKIP {skip} LIMIT {limit} RETURN n"
                        if label == "NO_LABEL"
                        else f"MATCH (n:{label}) SKIP {{skip}} LIMIT {{limit}} RETURN n"
                    )
                    tasks = [
                        loop.run_in_executor(
                            thread_pool,
                            self._fetch_nodes_chunk,
                            query_template.format(
                                skip=skip, limit=int(self.chunk_size)
                            ),
                        )
                        for skip in range(0, count, int(self.chunk_size))
                    ]
                chunks = await asyncio.gather(*tasks)
            except Exception as e:
                log.exception(
                    "Error fetching nodes for label '%s': %s from Neo4j", label, e
                )
                sys.exit(1)

            try:
                # Flatten chunks and assign new IDs starting from first_id
                all_data = [
                    {"id": idx + first_id, "properties": item}
                    for sublist in chunks
                    for idx, item in enumerate(sublist)
                ]
                # Build ID mapping for later edge export
                self.id_maps[label] = {}
                for item in all_data:
                    props = item.get("properties")
                    if isinstance(props, dict) and "_elementid" in props:
                        raw_id = item.get("id")
                        if isinstance(raw_id, int):
                            id_value = raw_id
                        elif isinstance(raw_id, str):
                            try:
                                id_value = int(raw_id)
                            except Exception:
                                log.exception(
                                    "Unable to convert id value %s to int.", raw_id
                                )
                                raise
                        else:
                            log.error("Unexpected type for id: %s", type(raw_id))
                            raise ValueError(f"Unsupported type for id: {type(raw_id)}")
                        self.id_maps[label][props["_elementid"]] = id_value
                file_path = await self.write_csv(label, "v", all_data)
                file_path = file_path.replace("\\", "\\\\")
                vertex_args[label] = {
                    "csv_path": file_path,
                    "next_val": str(len(all_data)),
                }
            except Exception as e:
                log.exception("Error exporting nodes for label '%s': %s", label, e)
        return vertex_args

    async def export_edges(
        self, thread_pool: concurrent.futures.Executor
    ) -> Dict[str, Dict[str, Any]]:
        """
        Export edges from Neo4j to CSV files and return a mapping for AGE import.
        """
        loop = asyncio.get_running_loop()
        edge_args: Dict[str, Dict[str, Any]] = {}
        types = self.get_relationship_types()
        for rel_type in types:
            try:
                # Create the edge label in AGE
                await self.create_label_type(label_type="edge", value=rel_type)
            except Exception as e:
                log.exception("Error creating edge type '%s': %s", rel_type, e)
                sys.exit(1)

            try:
                first_id = await self.get_first_id(self.graph_name, rel_type)
            except Exception as e:
                log.exception(
                    "Error getting first ID for edge type '%s': %s", rel_type, e
                )
                sys.exit(1)
            count = self._count_edges(rel_type)
            if self.trial:
                count = min(count, self.no_of_edges_trial)
                limit = min(int(self.chunk_size), count)
            else:
                limit = int(self.chunk_size)
            log.info("Exporting %d edges for relationship type '%s'.", count, rel_type)
            try:
                tasks = [
                    loop.run_in_executor(
                        thread_pool,
                        self._fetch_edge_chunk,
                        rel_type,
                        skip,
                        limit,
                    )
                    for skip in range(0, count, int(self.chunk_size))
                ]
                chunks = await asyncio.gather(*tasks)
                all_data = [
                    {
                        "id": idx + first_id,
                        "start_id": self.id_maps[item["start_vertex_type"]][
                            item["start_id"]
                        ],
                        "end_id": self.id_maps[item["end_vertex_type"]][item["end_id"]],
                        "properties": {
                            k: v
                            for k, v in item.items()
                            if k
                            not in [
                                "start_vertex_type",
                                "start_id",
                                "end_vertex_type",
                                "end_id",
                            ]
                        },
                    }
                    for sublist in chunks
                    for idx, item in enumerate(sublist)
                ]
                file_path = await self.write_csv(rel_type, "e", all_data)
                file_path = file_path.replace("\\", "\\\\")
                edge_args[rel_type] = {
                    "csv_path": file_path,
                    "next_val": str(len(all_data)),
                }
            except KeyError:
                log.error("Edge '%s' must include incomplete data.", rel_type)
            except Exception as e:
                log.exception("Error exporting edges for '%s': %s", rel_type, e)
        return edge_args

    async def list_nodes(self) -> None:
        """
        In trial mode, list nodes per relationship type for sampling.
        """
        loop = asyncio.get_running_loop()
        with concurrent.futures.ThreadPoolExecutor() as thread_pool:
            attempts = 0
            while True:
                attempts += 1
                try:
                    types = self.get_relationship_types()
                    break
                except Exception as e:
                    if attempts >= self.max_attempts:
                        log.error("Max attempts reached for getting relationship types")
                        sys.exit(1)
                    log.exception("Error getting relationship types: %s", e)
                    await asyncio.sleep(self.retry_delay)
            for rel_type in types:
                try:
                    count = min(self._count_edges(rel_type), self.no_of_edges_trial)
                    limit = min(int(self.chunk_size), count)
                    log.info("Listing nodes for relationship type '%s'.", rel_type)
                    tasks = [
                        loop.run_in_executor(
                            thread_pool,
                            self._fetch_edge_chunk,
                            rel_type,
                            skip,
                            limit,
                        )
                        for skip in range(0, count, limit)
                    ]
                    chunks = await asyncio.gather(*tasks)
                    all_data = [item for sublist in chunks for item in sublist]
                    nodes_by_label: Dict[str, List[str]] = {}
                    for record in all_data:
                        nodes_by_label.setdefault(
                            record["start_vertex_type"], []
                        ).append(record["start_id"])
                        nodes_by_label.setdefault(record["end_vertex_type"], []).append(
                            record["end_id"]
                        )
                    self.trial_nodes_by_label[rel_type] = nodes_by_label
                except Exception as e:
                    log.exception("Error listing nodes for '%s': %s", rel_type, e)

    async def export(self):
        """
        Main export function to process both nodes and edges.
        """
        thread_pool = concurrent.futures.ThreadPoolExecutor()
        attempts = 0
        while True:
            attempts += 1
            try:
                await self.set_up_graph(graph_name=self.graph_name, create_graph=True)
                break
            except Exception as e:
                if attempts >= self.max_attempts:
                    log.error("Max attempts reached for creating graph")
                    sys.exit(1)
                log.exception("Error creating graph: %s", e)
                await asyncio.sleep(self.retry_delay)
        try:
            if self.trial:
                await self.list_nodes()
            nodes_args = await self.export_nodes(thread_pool)
            if not nodes_args:
                log.error("No nodes exported.\nDoes the graph contain nodes?")
                sys.exit(1)
            edges_args = await self.export_edges(thread_pool)
            if not edges_args:
                log.error("No edges exported.\nDoes the graph contain edges?")
                sys.exit(1)
        except Exception as e:
            log.exception("Error during export process: %s", e)
            raise
        finally:
            thread_pool.shutdown()
        self.driver.close()

        self.vertices = nodes_args
        self.edges = edges_args
